from PySide6.QtCore import QTimer, Signal, QObject
from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QDialog,
    QCheckBox,
    QDialogButtonBox,
    QSpinBox,
    QLabel,
    QLineEdit,
)
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table
from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.ui_widgets.warn_popup import WarnPopup
from nomad_camels.utility import variables_handling


class Watchdog(QObject):
    condition_met = Signal(QObject)

    def __init__(
        self,
        channels=None,
        condition="",
        execute_at_condition="",
        read_periodic=False,
        read_timer=10,
        active=True,
        name="watchdog",
    ):
        super().__init__()
        self.name = name
        self.channels = channels or []
        self.condition = condition
        self.execute_at_condition = execute_at_condition
        self.read_periodic = read_periodic
        self.read_timer = read_timer
        self.active = active
        self.timer = QTimer()
        self.eva = None
        self.timer.timeout.connect(self.read)
        self.subscriptions = {}
        self.was_triggered = False
        self.plots = []
        self.ophyd_channels = []
        if self.read_periodic:
            self.timer.start(self.read_timer * 1000)

    def get_definition(self):
        return {
            "channels": self.channels,
            "condition": self.condition,
            "execute_at_condition": self.execute_at_condition,
            "read_periodic": self.read_periodic,
            "read_timer": self.read_timer,
            "active": self.active,
            "name": self.name,
        }

    def read(self):
        for channel in self.ophyd_channels:
            channel.read()

    def get_device_list(self):
        """Goes through the cannels and returns a list of all needed devices"""
        devices = []
        for channel in self.channels:
            chan = variables_handling.channels[channel]
            if chan.device not in devices:
                devices.append(chan.device)
        return devices

    def remove_device(self, device_name, ophyd_device):
        """Unsubscribes the respective channels corresponding to the device"""
        for channel in self.channels:
            try:
                chan = variables_handling.channels[channel]
                if chan.device == device_name:
                    ophyd_channel = getattr(ophyd_device, chan.name.split(".")[-1])
                    ophyd_channel.unsubscribe(self.subscriptions[channel])
                    self.subscriptions.pop(channel)
                    self.ophyd_channels.remove(ophyd_channel)
                if channel in self.eva.namespace:
                    self.eva.namespace.pop(channel)
            except:
                pass

    def add_device(self, device_name, ophyd_device):
        """Subscribes the respective channels corresponding to the device"""
        if not self.active:
            return
        for channel in self.channels:
            print(channel)
            if channel in self.subscriptions:
                continue
            chan = variables_handling.channels[channel]
            if chan.device == device_name:
                ophyd_channel = getattr(ophyd_device, chan.name.split(".")[-1])
                sub = ophyd_channel.subscribe(self.callback)
                self.subscriptions[channel] = sub
                self.ophyd_channels.append(ophyd_channel)

    def callback(self, value, **kwargs):
        if not self.active:
            return
        if "obj" in kwargs and hasattr(kwargs["obj"], "name"):
            self.eva.namespace[kwargs["obj"].name] = value
        try:
            condition = self.eva.eval(self.condition)
        except:
            print(f'Evaluating condition failed for watchdog "{self.name}"!')
            return
        if condition:
            if not self.was_triggered:
                self.condition_met.emit(self)
            self.was_triggered = True

    def update_settings(self):
        if self.timer.isActive():
            if not self.read_periodic:
                self.timer.stop()
            else:
                self.timer.setInterval(self.read_timer * 1000)
        elif self.read_periodic:
            self.timer.start(self.read_timer * 1000)


class Watchdog_Definer(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.watchdogs = dict(variables_handling.watchdogs)
        tabledata = {"Name": list(self.watchdogs.keys())}
        self.watchdog_table = AddRemoveTable(
            headerLabels=["Name"], tableData=tabledata, editables=[]
        )
        self.last_item = None

        self.watchdog_view = Watchdog_View(self, self.watchdogs)

        self.dialog_button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.dialog_button_box.accepted.connect(self.accept)
        self.dialog_button_box.rejected.connect(self.reject)

        self.setLayout(QGridLayout())
        self.layout().addWidget(self.watchdog_table, 0, 0)
        self.layout().addWidget(self.watchdog_view, 0, 1)
        self.layout().addWidget(self.dialog_button_box, 5, 0, 1, 2)

        self.watchdog_table.table.selectionModel().selectionChanged.connect(
            self.update_watchdog_view
        )

    def update_watchdogs(self):
        current_watchers = []
        for i in range(self.watchdog_table.table_model.rowCount()):
            watchdog = self.watchdog_table.table_model.item(i, 0).text()
            current_watchers.append(watchdog)
            if watchdog not in self.watchdogs:
                self.watchdogs[watchdog] = Watchdog()
        removes = []
        for watcher in self.watchdogs:
            if watcher not in current_watchers:
                removes.append(watcher)
        for remove in removes:
            self.watchdogs.pop(remove)

    def update_watchdog_view(self):
        if self.last_item is not None:
            try:
                old_name = self.last_item.text()
                self.watchdogs.pop(old_name)
                self.last_item.setText(self.watchdog_view.name_box.text())
                watchdog = self.watchdog_view.get_data()
                self.watchdogs[watchdog.name] = watchdog
            except RuntimeError:
                pass
        indexes = self.watchdog_table.table.selectionModel().selectedIndexes()
        if len(indexes) == 0:
            return
        self.update_watchdogs()
        index = indexes[0]
        self.last_item = self.watchdog_table.table_model.itemFromIndex(index)
        selected_watchdog = self.last_item.text()
        self.watchdog_view.update_watchdog(selected_watchdog)
        self.watchdog_table.table.resizeColumnsToContents()

    def accept(self):
        self.update_watchdogs()
        self.update_watchdog_view()
        variables_handling.watchdogs = self.watchdogs
        WarnPopup(
            self,
            "Watchdogs updated. Some functions might not work correclty right away. Restart CAMELS to make sure everything is working correctly.\n\nWatchdogs only read values of instruments that are being used by protocols or manual controls.\n(Further, channels that can only be read only work well if the instrument driver uses the Signals from CAMELS.)",
            "restart recommended, information",
            True,
        )
        super().accept()


class Watchdog_View(QWidget):
    def __init__(self, parent=None, watchdogs=None):
        super().__init__(parent)
        self.watchdog = None
        self.watchdogs = watchdogs
        self.condition_box = Variable_Box()
        self.channels_table = Channels_Check_Table(
            self, ["use", "channel"], title="connected channels"
        )

        self.label_protocol = QLabel("Execute at condition")
        self.protocol_selection = Path_Button_Edit(
            self,
            default_dir=variables_handling.preferences["py_files_path"],
            file_extension="*.cprot",
        )

        self.read_check = QCheckBox("force periodic read")
        self.read_timer = QSpinBox()
        self.label_name = QLabel("Name")
        self.name_box = QLineEdit()

        self.read_check.stateChanged.connect(self.read_check_changed)
        self.read_check.setChecked(False)
        self.read_timer.setValue(10)

        self.checkbox_active = QCheckBox("Watchdog Active")

        self.setLayout(QGridLayout())

        self.widgets = [
            self.condition_box,
            self.channels_table,
            self.protocol_selection,
            self.read_check,
            self.read_timer,
            self.label_name,
            self.name_box,
            self.checkbox_active,
        ]

        self.layout().addWidget(self.label_name, 0, 0)
        self.layout().addWidget(self.name_box, 0, 1)
        self.layout().addWidget(self.checkbox_active, 1, 0, 1, 2)
        self.layout().addWidget(self.condition_box, 2, 0, 1, 2)
        self.layout().addWidget(self.read_check, 3, 0)
        self.layout().addWidget(self.read_timer, 3, 1)
        self.layout().addWidget(self.channels_table, 4, 0, 1, 2)
        self.layout().addWidget(self.protocol_selection, 5, 0, 1, 2)

        self.update_watchdog(None)

    def update_watchdog(self, watchdog_name):
        if not watchdog_name in self.watchdogs:
            for widget in self.widgets:
                widget.setHidden(True)
            return
        for widget in self.widgets:
            widget.setHidden(False)
        self.watchdog = self.watchdogs[watchdog_name]
        self.condition_box.setText(self.watchdog.condition)
        self.channels_table.info_dict = {"channel": self.watchdog.channels}
        self.channels_table.build_channels_table()
        self.read_check.setChecked(self.watchdog.read_periodic)
        self.read_timer.setValue(self.watchdog.read_timer)
        self.checkbox_active.setChecked(self.watchdog.active)
        self.name_box.setText(self.watchdog.name)
        self.protocol_selection.set_path(self.watchdog.execute_at_condition)

    def read_check_changed(self):
        self.read_timer.setEnabled(self.read_check.isChecked())

    def get_data(self):
        if self.watchdog is None:
            return
        self.watchdog.condition = self.condition_box.text()
        self.watchdog.channels = self.channels_table.update_info()
        self.watchdog.channels = self.channels_table.info_dict["channel"]
        self.watchdog.read_periodic = self.read_check.isChecked()
        self.watchdog.read_timer = self.read_timer.value()
        self.watchdog.active = self.checkbox_active.isChecked()
        self.watchdog.name = self.name_box.text()
        self.watchdog.execute_at_condition = self.protocol_selection.get_path()
        self.watchdog.update_settings()
        return self.watchdog



import pytest
from nomad_camels.utility import variables_handling
from nomad_camels.main_classes.measurement_channel import (
    Measurement_Channel,
    from_pv_name,
)


@pytest.fixture
def sample_channel():
    return Measurement_Channel(
        name="sample.name",
        output=True,
        device="sample_device",
        metadata={"key1": "value1", "key2": "value2"},
    )


def test_measurement_channel_init(sample_channel):
    assert sample_channel.name == "sample.name"
    assert sample_channel.output is True
    assert sample_channel.device == "sample_device"
    assert sample_channel.metadata == {"key1": "value1", "key2": "value2"}


def test_get_bluesky_name(sample_channel):
    expected_bluesky_name = "sample_device.name"
    assert sample_channel.get_bluesky_name() == expected_bluesky_name


def test_get_pv_name(sample_channel, monkeypatch):
    monkeypatch.setattr(variables_handling, "preset", "TEST")
    expected_pv_name = "TEST:sample_device:name"
    assert sample_channel.get_pv_name() == expected_pv_name


def test_get_meta_str(sample_channel):
    expected_meta_str = "key1: value1\nkey2: value2"
    assert sample_channel.get_meta_str() == expected_meta_str


def test_from_pv_name():
    pv_name = "TEST:sample_device:name"
    expected_channel_name = "sample_device_name"
    assert from_pv_name(pv_name) == expected_channel_name



from ophyd import Signal, SignalRO, Device, Kind
import inspect
import time


class Custom_Function_Signal(Signal):
    """Overwrites ophyd's Signal to add a simple python function that's called
    when calling `put`, `trigger` or `get`.

    Attributes
    ----------
    put_function : callable
        Called when the Signal's `put` method is called.

    read_function : callable
        Called when the Signal's `get` method is called.

    trigger_function : callable
        Called when the Signal's `trigger` method is called.
    """

    def __init__(
        self,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        put_function=None,
        read_function=None,
        trigger_function=None,
        retry_on_error=0,
        error_retry_function=None,
        force_sequential=None,
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.put_function = put_function
        self.read_function = read_function
        self.trigger_function = trigger_function
        self.retry_on_error = retry_on_error
        self.error_retry_function = error_retry_function
        self.force_sequential = force_sequential

    def put(self, value, *, timestamp=None, force=False, metadata=None, **kwargs):
        """
        Overwrites Signal's `put` to add the defined `put_function`.
        For further information see ophyd's documentation.
        """
        if self.put_function:
            # If the put_function requires the instance of the class, pass it as the first argument
            # The instance is the signal, while the parent is the Device class
            # For dynamically created signals often a call to self.parent is required
            parent = False
            if self.force_sequential is not None:
                parent = self.parent if self.force_sequential else False
            elif getattr(self.parent, "force_sequential", None):
                parent = self.parent
            if (
                self.put_function
                and inspect.getfullargspec(self.put_function).args
                and inspect.getfullargspec(self.put_function).args[0]
                == "_self_instance"
            ):
                retry_function(
                    self.put_function(self, value),
                    self.retry_on_error,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
            else:
                retry_function(
                    self.put_function,
                    self.retry_on_error,
                    value,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
        super().put(
            value, timestamp=timestamp, force=force, metadata=metadata, **kwargs
        )

    def get(self):
        """
        Overwrites Signal's `get` to add the defined `read_function`.
        For further information see ophyd's documentation.
        """
        old_value = self._readback
        if self.read_function:
            # If the read_function requires the instance of the class, pass it as the first argument
            # The instance is the signal, while the parent is the Device class
            # For dynamically created signals often a call to self.parent is required
            parent = False
            if self.force_sequential is not None:
                parent = self.parent if self.force_sequential else False
            elif getattr(self.parent, "force_sequential", None):
                parent = self.parent
            if (
                self.read_function
                and inspect.getfullargspec(self.read_function).args
                and inspect.getfullargspec(self.read_function).args[0]
                == "_self_instance"
            ):
                self._readback = retry_function(
                    self.read_function(self),
                    self.retry_on_error,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
            else:
                self._readback = retry_function(
                    self.read_function,
                    self.retry_on_error,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
        self._run_subs(
            sub_type=self.SUB_VALUE,
            old_value=old_value,
            value=self._readback,
            timestamp=time.time(),
        )
        return super().get()

    def trigger(self):
        """
        Overwrites Signal's `trigger` to add the defined `trigger_function`.
        For further information see ophyd's documentation.
        """
        if self.trigger_function:
            retry_function(
                self.trigger_function,
                self.retry_on_error,
                error_retry_function=self.error_retry_function,
            )
        return super().trigger()

    def describe(self):
        """Overwrites describe to add 'Custom Function' as 'source'."""
        info = super().describe()
        info[self.name]["source"] = "Custom Function"
        return info

    def read_configuration(self):
        if self._readback is not None:
            value = self._readback
        else:
            value = self.get()
        return {self.name: {"value": value, "timestamp": self.timestamp}}


def retry_function(
    func, retries: int, *args, error_retry_function=None, parent=None, **kwargs
):
    """
    This function attempts to execute a given function multiple times until it succeeds or the maximum number of retries is reached.
    If a parent object is provided and it is currently reading, the function waits until the parent is no longer reading before attempting to execute the function.

    Parameters:
    func (callable): The function to be executed.
    retries (int): The maximum number of times to retry executing the function.
    *args: Variable length argument list for the function to be executed.
    error_retry_function (callable, optional): A function to be called when an error occurs. Defaults to None.
    parent (object, optional): An object (the instrument class) that the function checks for a currently_reading attribute. If currently_reading is True, the function waits until it is False before executing. Defaults to None.
    **kwargs: Arbitrary keyword arguments for the function to be executed.

    Returns:
    The return value of the function to be executed.

    Raises:
    Exception: If the function fails to execute after the specified number of retries, an exception is raised with details of the last exception encountered.
    """
    if parent:
        while parent.currently_reading:
            time.sleep(0.001)  # wait for 1 ms
        parent.currently_reading = True
    excs = []
    for i in range(retries + 1):
        try:
            result = func(*args, **kwargs)
            if parent:
                parent.currently_reading = False
            return result
        except Exception as e:
            excs.append(e)
            if error_retry_function:
                error_retry_function(e)
    if parent:
        parent.currently_reading = False
    raise Exception(
        f"Failed to execute function {func} after {retries} retries.\nLast exception: {excs[-1]}\nFirst exception: {excs[0]}"
    ) from excs[-1]


class Custom_Function_SignalRO(SignalRO):
    """Overwrites ophyd's SignalRO to add a simple python function that's called
    when calling, `trigger` or `get`.

    Attributes
    ----------
    read_function : callable
        Called when the Signal's `get` method is called.

    trigger_function : callable
        Called when the Signal's `trigger` method is called.
    """

    def __init__(
        self,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        read_function=None,
        trigger_function=None,
        retry_on_error=0,
        error_retry_function=None,
        force_sequential=None,
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.read_function = read_function
        self.trigger_function = trigger_function
        self.retry_on_error = retry_on_error
        self.error_retry_function = error_retry_function
        self.force_sequential = force_sequential

    def get(self):
        """
        Overwrites SignalRO's `get` to add the defined `read_function`.
        For further information see ophyd's documentation.
        """
        old_value = self._readback
        if self.read_function:
            # If the read_function requires the instance of the class, pass it as the first argument
            # The instance is the signal, while the parent is the Device class
            # For dynamically created signals often a call to self.parent is required
            parent = False
            if self.force_sequential is not None:
                parent = self.parent if self.force_sequential else False
            elif getattr(self.parent, "force_sequential", None):
                parent = self.parent
            if (
                self.read_function
                and inspect.getfullargspec(self.read_function).args
                and inspect.getfullargspec(self.read_function).args[0]
                == "_self_instance"
            ):
                self._readback = retry_function(
                    self.read_function(self),
                    self.retry_on_error,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
            else:
                self._readback = retry_function(
                    self.read_function,
                    self.retry_on_error,
                    error_retry_function=self.error_retry_function,
                    # parent is required if you want to force sequential reading
                    # parent is False if self.parent.force_sequential is False
                    # or if self.parent does not have a force_sequential attribute
                    parent=parent,
                )
        self._run_subs(
            sub_type=self.SUB_VALUE,
            old_value=old_value,
            value=self._readback,
            timestamp=time.time(),
        )
        return super().get()

    def trigger(self):
        """
        Overwrites SignalRO's `trigger` to add the defined `trigger_function`.
        For further information see ophyd's documentation.
        """
        if self.trigger_function:
            retry_function(
                self.trigger_function,
                self.retry_on_error,
                error_retry_function=self.error_retry_function,
            )
        return super().trigger()

    def describe(self):
        """Overwrites describe to add 'Custom Function' as 'source'."""
        info = super().describe()
        info[self.name]["source"] = "Custom Function"
        return info

    def read_configuration(self):
        if self._readback is not None and not (self.kind in ['config', Kind.config] and self.read_function) :
            value = self._readback
        else:
            value = self.get()
        return {self.name: {"value": value, "timestamp": self.timestamp}}


class Sequential_Device(Device):
    def __init__(
        self, force_sequential=False, currently_reading=False, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.force_sequential = force_sequential
        self.currently_reading = currently_reading



import os

from PySide6.QtWidgets import QLabel, QComboBox, QCheckBox

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config
from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.utility import variables_handling, load_save_functions
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.bluesky_handling import protocol_builder, builder_helper_functions
from nomad_camels.bluesky_handling import protocol_builder


class Run_Subprotocol(Loop_Step):
    """
    With this step, one may select another protocol to run inside the main one.

    Attributes
    ----------
    prot_path : str, path
        Path to the file of the subprotocol.
    vars_in : dict
        Variables of the subprotocol's namespace and the values they should get
        before the subprotocol is run. This can be used to e.g. give the
        subprotocol a new value for each run inside a loop.
    vars_out : dict
        Variables of the subprotocol and the name in the main protocol's
        namespace where they should be put. This can be used to e.g. store some
        value determined by the subprotocol for later use in the main protocol.
    data_output : str
        Whether the data is put into its own stream('sub-stream') or the primary
        stream ('main-stream').
    own_plots : bool
        If True, the plots specified by the protocol will also be shown.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Run Subprotocol"
        if step_info is None:
            step_info = {}
        self.prot_path = step_info["prot_path"] if "prot_path" in step_info else ""
        self.vars_in = step_info["vars_in"] if "vars_in" in step_info else {}
        self.vars_out = step_info["vars_out"] if "vars_out" in step_info else {}
        self.data_output = (
            step_info["data_output"] if "data_output" in step_info else "sub-stream"
        )
        self.own_plots = step_info["own_plots"] if "own_plots" in step_info else True

    def get_protocol_string(self, n_tabs=1):
        """Overwrites the signal for the progressbar and the number of steps in
        the subprotocol's module. Evaluates the input variables, then writes
        them into the subprotocol's namespace and starts the subprotocol's
        _plan_inner function. Afterwards the output variables are written to the
        main namespace."""
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += protocol_builder.sub_protocol_string(
            protocol_path=self.prot_path,
            n_tabs=n_tabs,
            variables_in=self.vars_in,
            variables_out=self.vars_out,
            data_output=self.data_output,
        )
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Specifies the name / path of the subprotocol."""
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]} - {self.prot_path}"
        return short_string

    def get_outer_string(self):
        """Imports the subprotocol as <protocol_name>_mod."""
        return protocol_builder.import_protocol_string(self.prot_path)
        # prot_name = os.path.basename(self.prot_path)[:-6]
        # py_file = f"{self.prot_path[:-6]}.py"
        # if not os.path.isfile(py_file):
        #     sub_protocol = load_save_functions.load_protocol(self.prot_path)
        #     protocol_builder.build_protocol(sub_protocol, py_file)
        # outer_string = f'spec = importlib.util.spec_from_file_location("{prot_name}", "{py_file}")\n'
        # outer_string += f"{prot_name}_mod = importlib.util.module_from_spec(spec)\n"
        # outer_string += f"sys.modules[spec.name] = {prot_name}_mod\n"
        # outer_string += f"spec.loader.exec_module({prot_name}_mod)\n"
        # return outer_string

    def get_add_main_string(self):
        """If using its own plots, adds them to the steps. In any case, the
        added steps from the subprotocol are added here as well."""
        return protocol_builder.make_plots_string_of_protocol(
            self.prot_path, self.own_plots, self.data_output, 1
        )
        # prot_name = os.path.basename(self.prot_path)[:-6]
        # add_main_string = ""
        # if self.own_plots:
        #     stream = f'"{prot_name}"'
        #     if self.data_output == "main stream":
        #         stream = '"primary"'
        #     add_main_string += builder_helper_functions.get_plot_add_string(
        #         prot_name, stream, True
        #     )
        # add_main_string += f'\treturner["{prot_name}_steps"] = {prot_name}_mod.steps_add_main(RE, devs)\n'
        # return add_main_string

    def update_used_devices(self):
        """Uses the devices that are used in the subprotocol."""
        sub_protocol = load_save_functions.load_protocol(self.prot_path)
        self.used_devices = sub_protocol.get_used_devices()

    def update_time_weight(self):
        """The time weight in the end is the weight of the subprotocol + 1."""
        super().update_time_weight()
        sub_protocol = load_save_functions.load_protocol(self.prot_path)
        self.time_weight += sub_protocol.get_total_steps()


class Run_Subprotocol_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Run_Subprotocol, parent=None):
        super().__init__(parent, loop_step)
        label = QLabel("Select Protocol:")
        self.loop_step = loop_step
        self.path_button = Path_Button_Edit(
            self,
            loop_step.prot_path,
            variables_handling.preferences["py_files_path"],
            file_extension="*.cprot",
        )
        self.sub_vars = {}
        self.load_sub_vars()
        headerLabels = ["Variable", "Value"]
        comboBoxes = {"Variable": self.sub_vars.keys()}
        self.input_table = AddRemoveTable(
            headerLabels=headerLabels,
            tableData=loop_step.vars_in,
            title="Variables In",
            comboBoxes=comboBoxes,
            checkstrings=[1],
        )
        headerLabels = ["Variable", "Write to name"]
        comboBoxes = {"Variable": self.sub_vars.keys()}
        self.output_table = AddRemoveTable(
            headerLabels=headerLabels,
            tableData=loop_step.vars_out,
            title="Variables Out",
            comboBoxes=comboBoxes,
        )

        label_data = QLabel("Data Output:")
        self.comboBox_data_output = QComboBox()
        output_types = ["sub-stream", "main stream"]  # , 'own file']
        self.comboBox_data_output.addItems(output_types)
        self.comboBox_data_output.setCurrentText(loop_step.data_output)

        self.checkBox_plots = QCheckBox("Use own plots")
        self.checkBox_plots.setChecked(loop_step.own_plots)

        self.layout().addWidget(label, 1, 0)
        self.layout().addWidget(self.path_button, 1, 1)
        self.layout().addWidget(label_data, 2, 0)
        self.layout().addWidget(self.comboBox_data_output, 2, 1)
        self.layout().addWidget(self.checkBox_plots, 3, 0, 1, 2)
        self.layout().addWidget(self.input_table, 5, 0, 1, 2)
        self.layout().addWidget(self.output_table, 6, 0, 1, 2)
        self.path_button.path_changed.connect(self.update_sub_vars)

    def update_sub_vars(self):
        """ """
        self.load_sub_vars()
        comboBoxes = {"Variable": self.sub_vars.keys()}
        self.input_table.comboBoxes = comboBoxes
        self.input_table.load_table_data()
        self.output_table.comboBoxes = comboBoxes
        self.output_table.load_table_data()

    def load_sub_vars(self):
        """ """
        self.sub_vars = {}
        prot_path = self.path_button.get_path()
        sub_protocol = load_save_functions.load_protocol(prot_path)
        if sub_protocol:
            self.sub_vars = sub_protocol.variables

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.loop_step.prot_path = self.path_button.get_path()
        self.loop_step.vars_in = self.input_table.update_table_data()
        self.loop_step.vars_out = self.output_table.update_table_data()
        self.loop_step.own_plots = self.checkBox_plots.isChecked()
        self.loop_step.data_output = self.comboBox_data_output.currentText()



from copy import deepcopy

from PySide6.QtWidgets import QWidget, QToolButton, QMenu, QMessageBox, QDialogButtonBox
from PySide6.QtCore import QItemSelectionModel, Qt, Signal
from PySide6.QtGui import (
    QStandardItemModel,
    QAction,
    QCloseEvent,
    QKeyEvent,
    QShortcut,
    QIcon,
    QUndoStack,
)

from nomad_camels.gui.protocol_view import Ui_Protocol_View
from nomad_camels.main_classes.protocol_class import (
    Measurement_Protocol,
    General_Protocol_Settings,
)
from nomad_camels.loop_steps import make_step_of_type
from nomad_camels.utility import variables_handling, treeView_functions
from nomad_camels.ui_widgets import drag_drop_tree_view
from nomad_camels.commands import change_sequence

from importlib import resources
from nomad_camels import graphics


loop_step_display_order = [
    "Read Channels",
    "Set Channels",
    "For Loop",
    "Simple Sweep",
    "Wait",
    "Run Subprotocol",
    "If",
    "Set Variables",
    "Execute Python File",
    "ND Sweep",
    "Change Device Config",
    "While Loop",
    "Gradient Descent",
    "API Call",
    "Set Value Popup",
    "Trigger Channels",
    "Prompt",
    "Call Function",
]

channel_action_list = [
    "Read Channels",
    "Set Channels",
    "Trigger Channels",
]
loop_action_list = [
    "For Loop",
    "While Loop",
    "Simple Sweep",
    "ND Sweep",
]
additional_action_list = [
    "Execute Python File",
    "API Call",
    "Change Device Config",
    "Gradient Descent",
    "Prompt",
    "Call Function",
    "Set Value Popup",
]


class Protocol_Config(Ui_Protocol_View, QWidget):
    """ """

    accepted = Signal(Measurement_Protocol)
    closing = Signal()

    def __init__(self, protocol=None, parent=None):
        super().__init__(parent=parent)
        self.old_name = None
        if protocol is None:
            protocol = Measurement_Protocol()
        else:
            protocol = deepcopy(protocol)
            self.old_name = protocol.name
        self.setupUi(self)
        self.setWindowTitle(f"{protocol.name} - Measurement Protocol - NOMAD CAMELS")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        self.configuration_main_widget.setHidden(True)
        self.general_settings = General_Protocol_Settings(protocol=protocol)
        self.meas_splitter.insertWidget(0, self.general_settings)

        class AddButton(QToolButton):
            def __init__(self, function, parent=None):
                super().__init__(parent)
                self.setText("+")
                self.function = function

            def mousePressEvent(self, event):
                if event.button() == Qt.LeftButton:
                    self.function(event.pos())
                else:
                    super().mousePressEvent(event)

        add_button = AddButton(self.sequence_right_click, self)
        stylesheet = self.toolButton_add_step.styleSheet()
        add_button.setStyleSheet(stylesheet)
        self.sequence_main_widget.layout().replaceWidget(
            self.toolButton_add_step, add_button
        )
        self.toolButton_add_step.deleteLater()
        self.toolButton_add_step = add_button
        self.protocol = protocol
        self.loop_step_configuration_widget = None

        self.add_actions = []
        self.device_actions = []
        self.is_accepted = False

        self.item_model_sequence = QStandardItemModel(0, 1)
        self.treeView_protocol_sequence = drag_drop_tree_view.Drag_Drop_TreeView()
        self.treeView_protocol_sequence.del_clicked.connect(self.remove_loop_step)
        self.sequence_main_widget.layout().addWidget(
            self.treeView_protocol_sequence, 5, 0, 1, 3
        )
        self.treeView_protocol_sequence.setModel(self.item_model_sequence)
        self.treeView_protocol_sequence.customContextMenuRequested.connect(
            self.sequence_right_click
        )
        self.treeView_protocol_sequence.dragdrop.connect(self.update_loop_step_order)

        self.pushButton_move_step_up.clicked.connect(
            lambda state: self.move_loop_step(-1, 0)
        )
        self.pushButton_move_step_down.clicked.connect(
            lambda state: self.move_loop_step(1, 0)
        )
        self.pushButton_move_step_in.clicked.connect(
            lambda state: self.move_loop_step(0, 1)
        )
        self.pushButton_move_step_out.clicked.connect(
            lambda state: self.move_loop_step(0, -1)
        )
        self.treeView_protocol_sequence.clicked.connect(
            lambda x: self.tree_click_sequence()
        )
        self.pushButton_remove_step.clicked.connect(
            lambda x: self.remove_loop_step(True)
        )
        self.general_settings.name_changed.connect(self.change_name)

        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.rejected.connect(self.close)
        self.buttonBox.accepted.connect(self.accept)

        self.update_add_step_actions()

        self.layout().addWidget(self.buttonBox, 15, 0)

        self.inside_function = False
        self.undo_stack = QUndoStack(self)
        # QShortcut('Ctrl+z', self).activated.connect(self.undo)
        # QShortcut('Ctrl+y', self).activated.connect(self.redo)

        QShortcut(
            "Ctrl+x", self.sequence_main_widget, context=Qt.WidgetWithChildrenShortcut
        ).activated.connect(self.cut_shortcut)
        QShortcut(
            "Ctrl+v", self.sequence_main_widget, context=Qt.WidgetWithChildrenShortcut
        ).activated.connect(self.paste_shortcut)
        QShortcut(
            "Ctrl+c", self.sequence_main_widget, context=Qt.WidgetWithChildrenShortcut
        ).activated.connect(self.copy_shortcut)
        QShortcut(
            "Ctrl+k", self.sequence_main_widget, context=Qt.WidgetWithChildrenShortcut
        ).activated.connect(self.comment_shortcut)

        self.build_protocol_sequence()
        self.check_movability()

    def update_add_step_actions(self):
        """Called when the devices change, updating the possible
        loopsteps to include new device steps.

        Parameters
        ----------

        Returns
        -------

        """
        self.add_actions.clear()
        self.device_actions.clear()
        for stp in sorted(
            make_step_of_type.step_type_config.keys(),
            key=lambda x: loop_step_display_order.index(x),
        ):
            action = QAction(stp)
            action.triggered.connect(lambda state=None, x=stp: self.add_loop_step(x))
            self.add_actions.append(action)
        for stp in make_step_of_type.get_device_steps():
            action = QAction(stp)
            action.triggered.connect(lambda state=None, x=stp: self.add_loop_step(x))
            self.device_actions.append(action)
        # self.toolButton_add_step.addActions(self.add_actions)
        # if self.device_actions:
        #     self.toolButton_add_step.addActions(self.device_actions)

    def tree_click_sequence(self):
        """Called when clicking the treeView_protocol_sequence."""
        # Preserve the current scroll position
        scroll_position = self.treeView_protocol_sequence.verticalScrollBar().value()
        self.update_loop_step_order()
        self.get_step_config()
        self.protocol.update_variables()
        self.configuration_main_widget.setHidden(False)
        config = None
        index = self.treeView_protocol_sequence.selectedIndexes()
        if not index:
            return
        index = index[0]
        dat = self.item_model_sequence.itemFromIndex(index).data()
        if dat is not None:
            step = self.protocol.loop_step_dict[dat]
            config = make_step_of_type.get_config(step)
            enable = step.step_type not in make_step_of_type.non_addables
            self.enable_step_move(enable)
            self.label_configuration.setText(f"Configuration: {step.full_name}")
        if config is not None:
            if self.loop_step_configuration_widget is not None:
                self.configuration_main_widget.layout().removeWidget(
                    self.loop_step_configuration_widget
                )
                self.loop_step_configuration_widget.deleteLater()
            self.loop_step_configuration_widget = config
            self.configuration_main_widget.layout().addWidget(
                self.loop_step_configuration_widget, 1, 0
            )
            self.loop_step_configuration_widget.name_changed.connect(
                self.change_step_name
            )
            self.loop_step_configuration_widget.active_changed.connect(
                self.change_step_name
            )
        self.check_movability()
        # Restore the scroll position
        self.treeView_protocol_sequence.verticalScrollBar().setValue(scroll_position)

    def build_protocol_sequence(self):
        """Shows / builds the protocol sequence in the treeView
        dependent on the loop_steps in the current_protocol.

        Parameters
        ----------

        Returns
        -------

        """
        ind_seq = self.treeView_protocol_sequence.selectedIndexes()
        sel_data = ""
        if ind_seq:
            sel_data = self.item_model_sequence.data(ind_seq[0])
        variables_handling.current_protocol = self.protocol
        variables_handling.protocol_variables = self.protocol.variables
        variables_handling.loop_step_variables = self.protocol.loop_step_variables
        self.item_model_sequence.clear()
        for loop_step in self.protocol.loop_steps:
            loop_step.append_to_model(self.item_model_sequence)
        self.treeView_protocol_sequence.expandAll()
        self.sequence_main_widget.setEnabled(True)
        if sel_data is not None:
            new_index = treeView_functions.getItemIndex(
                self.item_model_sequence, sel_data
            )
            if new_index:
                self.treeView_protocol_sequence.selectionModel().select(
                    new_index, QItemSelectionModel.Select
                )

    def enable_step_move(self, enable):
        """

        Parameters
        ----------
        enable :


        Returns
        -------

        """
        self.pushButton_move_step_in.setEnabled(enable)
        self.pushButton_move_step_out.setEnabled(enable)
        self.pushButton_move_step_up.setEnabled(enable)
        self.pushButton_move_step_down.setEnabled(enable)

    def change_step_name(self):
        """Called when a loop_step changes its name, then updates the
        shown sequence, and also the protocol-data.

        Parameters
        ----------

        Returns
        -------

        """
        self.build_protocol_sequence()
        self.update_loop_step_order()

    def get_step_config(self):
        """Updates the data in the currently-to-configure loop_step."""
        if self.loop_step_configuration_widget is not None:
            self.loop_step_configuration_widget.update_step_config()

    def sequence_right_click(self, pos):
        """Opens a specific Menu on right click in the protocol-sequence.
        If selection is not on a loop_step, it consists only of Add Step,
        otherwise it consists of Delete Step.

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        # TODO other actions
        # TODO more beautiful?
        menu = QMenu()
        inds = self.treeView_protocol_sequence.selectedIndexes()
        if inds:
            item = self.item_model_sequence.itemFromIndex(inds[0])
            below_actions = []
            above_actions = []
            into_actions = []
            replace_actions = []
            row = inds[0].row()
            parent = item.parent()
            if parent is not None:
                parent = parent.data()
            for stp in sorted(
                make_step_of_type.step_type_config.keys(),
                key=lambda x: loop_step_display_order.index(x),
            ):
                action = QAction(stp)
                action_a = QAction(stp)
                action_in = QAction(stp)
                action.triggered.connect(
                    lambda state=None, x=stp, y=row + 1, z=parent: self.add_loop_step(
                        x, y, z
                    )
                )
                action_a.triggered.connect(
                    lambda state=None, x=stp, y=row, z=parent: self.add_loop_step(
                        x, y, z
                    )
                )
                action_in.triggered.connect(
                    lambda state=None, x=stp, y=-1, z=item.data(): self.add_loop_step(
                        x, y, z
                    )
                )
                below_actions.append(action)
                above_actions.append(action_a)
                into_actions.append(action_in)
                if (
                    not self.protocol.loop_step_dict[item.data()].has_children
                    or not self.protocol.loop_step_dict[item.data()].children
                    or stp in make_step_of_type.steps_with_children
                ):
                    action_replace = QAction(stp)
                    action_replace.triggered.connect(
                        lambda state=None, x=stp, y=row, z=parent: self.replace_loop_step(
                            x, y, z
                        )
                    )
                    replace_actions.append(action_replace)
            device_actions = []
            device_actions_a = []
            device_actions_in = []
            for stp in make_step_of_type.get_device_steps():
                action = QAction(stp)
                action_a = QAction(stp)
                action_in = QAction(stp)
                action.triggered.connect(
                    lambda state=None, x=stp, y=row + 1, z=parent: self.add_loop_step(
                        x, y, z
                    )
                )
                action_a.triggered.connect(
                    lambda state=None, x=stp, y=row, z=parent: self.add_loop_step(
                        x, y, z
                    )
                )
                action_in.triggered.connect(
                    lambda state=None, x=stp, y=-1, z=item.data(): self.add_loop_step(
                        x, y, z
                    )
                )
                device_actions.append(action)
                device_actions_a.append(action_a)
                device_actions_in.append(action_in)

            # -------------- Above actions -----------------------------
            insert_above_menu = QMenu("Insert Above")
            insert_above_menu_channels = QMenu("Channels")
            insert_above_menu_loops = QMenu("Loops")
            insert_above_menu_additional = QMenu("Advanced")
            # Filter out specific steps with a list of the .text of the QAction and then remove them from the list
            duplicate_actions = above_actions[:]
            for action in above_actions:
                if action.text() in channel_action_list:
                    insert_above_menu_channels.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in loop_action_list:
                    insert_above_menu_loops.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in additional_action_list:
                    insert_above_menu_additional.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
            insert_above_menu.addMenu(insert_above_menu_channels)
            insert_above_menu.addMenu(insert_above_menu_loops)
            if len(duplicate_actions) > 0:
                insert_above_menu.addActions(duplicate_actions)
            insert_above_menu.addMenu(insert_above_menu_additional)

            # -------------- Below actions -----------------------------
            insert_below_menu = QMenu("Insert Below")
            insert_below_menu_channels = QMenu("Channels")
            insert_below_menu_loops = QMenu("Loops")
            insert_below_menu_additional = QMenu("Advanced")
            # Filter out specific steps with a list of the .text of the QAction and then remove them from the list for the below menu
            duplicate_actions = below_actions[:]
            for action in below_actions:
                if action.text() in channel_action_list:
                    insert_below_menu_channels.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in loop_action_list:
                    insert_below_menu_loops.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in additional_action_list:
                    insert_below_menu_additional.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
            insert_below_menu.addMenu(insert_below_menu_channels)
            insert_below_menu.addMenu(insert_below_menu_loops)
            if len(duplicate_actions) > 0:
                insert_below_menu.addActions(duplicate_actions)
            insert_below_menu.addMenu(insert_below_menu_additional)
            # -------------- Add in actions -----------------------------
            # insert_below_menu.addActions(below_actions)
            if device_actions:
                insert_above_menu.addSeparator()
                insert_above_menu.addActions(device_actions_a)
                insert_below_menu.addSeparator()
                insert_below_menu.addActions(device_actions)
            if self.protocol.loop_step_dict[item.data()].has_children:
                add_in_menu = QMenu("Add Into")
                add_in_menu_channels = QMenu("Channels")
                add_in_menu_loops = QMenu("Loops")
                add_in_menu_additional = QMenu("Advanced")
                # Filter out specific steps with a list of the .text of the QAction and then remove them from the list
                duplicate_actions = into_actions[:]
                for action in into_actions:
                    if action.text() in channel_action_list:
                        add_in_menu_channels.addAction(action)
                        # Remove from the list
                        duplicate_actions.remove(action)
                    elif action.text() in loop_action_list:
                        add_in_menu_loops.addAction(action)
                        # Remove from the list
                        duplicate_actions.remove(action)
                    elif action.text() in additional_action_list:
                        add_in_menu_additional.addAction(action)
                        # Remove from the list
                        duplicate_actions.remove(action)
                add_in_menu.addMenu(add_in_menu_channels)
                add_in_menu.addMenu(add_in_menu_loops)
                if len(duplicate_actions) > 0:
                    add_in_menu.addActions(duplicate_actions)
                add_in_menu.addMenu(add_in_menu_additional)
                menu.addMenu(add_in_menu)
                if device_actions:
                    add_in_menu.addSeparator()
                    add_in_menu.addActions(device_actions_in)
            replace_menu = QMenu("Replace with")
            replace_menu.addActions(replace_actions)
            menu.addMenu(insert_above_menu)
            menu.addMenu(insert_below_menu)
            menu.addMenu(replace_menu)
            menu.addSeparator()
            cut_action = QAction("Cut\tCtrl + X")
            cut_action.triggered.connect(
                lambda state=None, x=item.data(): self.cut_loop_step(x)
            )
            copy_action = QAction("Copy\tCtrl + C")
            copy_action.triggered.connect(
                lambda state=None, x=item.data(): self.copy_loop_step(x)
            )
            paste_menu = QMenu("Paste\tCtrl + V")
            if variables_handling.copied_step is not None:
                paste_above = QAction("Paste Above")
                paste_above.triggered.connect(
                    lambda state=None, x=True, y=row, z=parent: self.add_loop_step(
                        copied_step=x, position=y, parent=z
                    )
                )
                paste_below = QAction("Paste Below")
                paste_below.triggered.connect(
                    lambda state=None, x=True, y=row + 1, z=parent: self.add_loop_step(
                        copied_step=x, position=y, parent=z
                    )
                )
                if self.protocol.loop_step_dict[item.data()].has_children:
                    paste_into = QAction("Paste Into")
                    paste_into.triggered.connect(
                        lambda state=None, x=True, y=-1, z=item.data(): self.add_loop_step(
                            copied_step=x, position=y, parent=z
                        )
                    )
                    paste_menu.addAction(paste_into)
                paste_menu.addActions([paste_above, paste_below])
            else:
                paste_menu.setEnabled(False)
            menu.addAction(cut_action)
            menu.addAction(copy_action)
            menu.addMenu(paste_menu)
            if (
                self.protocol.loop_step_dict[item.data()].step_type
                not in make_step_of_type.non_addables
            ):
                del_action = QAction("Delete Step\tDel")
                del_action.triggered.connect(lambda x: self.remove_loop_step(True))
                menu.addAction(del_action)
            menu.addSeparator()
            if item.text().startswith("# "):
                comment_action = QAction("Enable step\tCtrl + K")
            else:
                comment_action = QAction("Disable step\tCtrl + K")
            comment_action.triggered.connect(
                lambda state=None, x=item.data(): self.comment_loop_step(x)
            )
            menu.addAction(comment_action)
        else:
            add_actions = []
            for stp in sorted(
                make_step_of_type.step_type_config,
                key=lambda x: loop_step_display_order.index(x),
            ):
                action = QAction(stp)
                action.triggered.connect(
                    lambda state=None, x=stp: self.add_loop_step(x)
                )
                add_actions.append(action)
            device_actions = []
            for stp in make_step_of_type.get_device_steps():
                action = QAction(stp)
                action.triggered.connect(
                    lambda state=None, x=stp: self.add_loop_step(x)
                )
                device_actions.append(action)
            add_menu = QMenu("Add Step")
            # -------------- Above actions -----------------------------
            add_menu_channels = QMenu("Channels")
            add_menu_loops = QMenu("Loops")
            add_menu_additional = QMenu("Advanced")
            # Filter out specific steps with a list of the .text of the QAction and then remove them from the list
            duplicate_actions = add_actions[:]
            for action in add_actions:
                if action.text() in channel_action_list:
                    add_menu_channels.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in loop_action_list:
                    add_menu_loops.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
                elif action.text() in additional_action_list:
                    add_menu_additional.addAction(action)
                    # Remove from the list
                    duplicate_actions.remove(action)
            add_menu.addMenu(add_menu_channels)
            add_menu.addMenu(add_menu_loops)
            if len(duplicate_actions) > 0:
                add_menu.addActions(duplicate_actions)
            add_menu.addMenu(add_menu_additional)
            if device_actions:
                add_menu.addSeparator()
                add_menu.addActions(device_actions)
            paste_action = QAction("Paste")
            if variables_handling.copied_step is not None:
                paste_action.triggered.connect(
                    lambda state=None, x=True, y=-1, z=None: self.add_loop_step(
                        copied_step=x, position=y, parent=z
                    )
                )
            else:
                paste_action.setEnabled(False)
            menu.addMenu(add_menu)
            menu.addAction(paste_action)
        menu.exec_(self.treeView_protocol_sequence.viewport().mapToGlobal(pos))

    def paste_shortcut(self):
        """ """
        if variables_handling.copied_step is None:
            return
        inds = self.treeView_protocol_sequence.selectedIndexes()
        if inds:
            ind = inds[0]
            item = self.item_model_sequence.itemFromIndex(ind)
            if self.protocol.loop_step_dict[item.data()].has_children:
                pos = -1
                parent = item.data()
            else:
                pos = ind.row() + 1
                if item.parent():
                    parent = item.parent().data()
                else:
                    parent = None
        else:
            pos = -1
            parent = None
        self.add_loop_step(copied_step=True, position=pos, parent=parent)

    def cut_shortcut(self):
        """ """
        inds = self.treeView_protocol_sequence.selectedIndexes()
        if not inds:
            return
        item = self.item_model_sequence.itemFromIndex(inds[0])
        self.cut_loop_step(item.data())

    def copy_shortcut(self):
        """ """
        inds = self.treeView_protocol_sequence.selectedIndexes()
        if not inds:
            return
        item = self.item_model_sequence.itemFromIndex(inds[0])
        self.copy_loop_step(item.data())

    def comment_shortcut(self):
        inds = self.treeView_protocol_sequence.selectedIndexes()
        if not inds:
            return
        item = self.item_model_sequence.itemFromIndex(inds[0])
        self.comment_loop_step(item.data())

    def cut_loop_step(self, step_name):
        """Copies the given step, then removes it.

        Parameters
        ----------
        step_name :


        Returns
        -------

        """
        self.copy_loop_step(step_name)
        self.remove_loop_step(ask=False)

    def copy_loop_step(self, step_name):
        """Makes a deepcopy of the given step and stores it in
        copied_loop_step.

        Parameters
        ----------
        step_name :


        Returns
        -------

        """
        variables_handling.copied_step = deepcopy(
            self.protocol.loop_step_dict[step_name]
        )

    def check_movability(self):
        """ """
        ind = self.treeView_protocol_sequence.selectedIndexes()
        if not ind:
            self.enable_step_move(False)
            self.pushButton_remove_step.setEnabled(False)
            return
        ind = ind[0]
        item = self.item_model_sequence.itemFromIndex(ind)
        self.pushButton_remove_step.setEnabled(True)
        parent = item.parent()
        if ind.row() > 0:
            self.pushButton_move_step_up.setEnabled(True)
            if parent is not None:
                above = parent.child(ind.row() - 1, 0)
            else:
                above = self.item_model_sequence.item(ind.row() - 1, 0)
            self.pushButton_move_step_in.setEnabled(True)
            if above is not None:
                step = self.protocol.loop_step_dict[above.data()]
                if step.step_type == "If":
                    above = above.child(above.rowCount() - 1, 0)
                if not self.protocol.loop_step_dict[above.data()].has_children:
                    self.pushButton_move_step_in.setEnabled(False)
        else:
            self.pushButton_move_step_up.setEnabled(False)
            self.pushButton_move_step_in.setEnabled(False)
        if ind.row() < self.item_model_sequence.rowCount() - 1:
            self.pushButton_move_step_down.setEnabled(True)
        else:
            self.pushButton_move_step_down.setEnabled(False)
        if parent is None:
            self.pushButton_move_step_out.setEnabled(False)
        else:
            self.pushButton_move_step_out.setEnabled(True)

    def move_loop_step(self, up_down=0, in_out=0):
        """Moves a loop_step up or down in the sequence. It can also be
        moved in or out (into the loop_step above, it if accepts children).

        Parameters
        ----------
        up_down :
             (Default value = 0)
        in_out :
             (Default value = 0)

        Returns
        -------


        """
        move_command = change_sequence.CommandMoveStep(
            self.treeView_protocol_sequence,
            self.item_model_sequence,
            up_down,
            in_out,
            self.protocol.loop_step_dict,
            self.update_loop_step_order,
        )
        self.undo_stack.push(move_command)
        self.tree_click_sequence()

    def comment_loop_step(self, step_name):
        step = self.protocol.loop_step_dict[step_name]
        step.is_active = not step.is_active
        self.build_protocol_sequence()

    def add_loop_step(self, step_type="", position=-1, parent=None, copied_step=False):
        """Add a loop_step of given step_type. Updates the current
        sequence into the protocol, then initializes the new step.

        Parameters
        ----------
        step_type :
             (Default value = '')
        position :
             (Default value = -1)
        parent :
             (Default value = None)
        copied_step :
             (Default value = False)

        Returns
        -------
        step : Loop_Step
            the newly added loop_step

        """

        self.update_loop_step_order()
        if copied_step:
            step = variables_handling.copied_step
        else:
            step = make_step_of_type.make_step(step_type, protocol=self.protocol)
        self.protocol.add_loop_step_rec(
            step,
            model=self.item_model_sequence,
            position=position,
            parent_step_name=parent,
        )
        self.build_protocol_sequence()
        if copied_step:
            self.copy_loop_step(variables_handling.copied_step.full_name)
        new_ind = treeView_functions.getItemIndex(
            self.item_model_sequence, step.full_name
        )
        self.treeView_protocol_sequence.selectionModel().select(
            new_ind, QItemSelectionModel.Select
        )
        self.tree_click_sequence()
        return step

    def remove_loop_step(self, ask=True):
        """After updating the loop_step order in the protocol, the
        selected loop step is deleted (if the messagebox is accepted).

        Parameters
        ----------
        ask :
             (Default value = True)

        Returns
        -------

        """
        self.update_loop_step_order()
        ind = self.treeView_protocol_sequence.selectedIndexes()[0]
        name = self.item_model_sequence.itemFromIndex(ind).data()
        if name is not None:
            remove_dialog = None
            if (
                self.protocol.loop_step_dict[name].step_type
                in make_step_of_type.non_addables
            ):
                return
            if ask:
                remove_dialog = QMessageBox.question(
                    self,
                    "Delete Step?",
                    f"Are you sure you want to delete the step {name}?",
                    QMessageBox.Yes | QMessageBox.No,
                )
            if not ask or remove_dialog == QMessageBox.Yes:
                self.protocol.remove_loop_step(name)
                self.build_protocol_sequence()
                self.check_movability()

    def replace_loop_step(self, step_type="", position=-1, parent=None):
        """ """
        ind = self.treeView_protocol_sequence.selectedIndexes()[0]
        children_names = treeView_functions.get_substeps(
            self.item_model_sequence.itemFromIndex(ind)
        )
        children_steps = []
        for child in children_names:
            children_steps.append(self.protocol.loop_step_dict[child[0]])
        self.remove_loop_step(ask=False)
        step = self.add_loop_step(step_type, position, parent)
        step.children = children_steps
        self.build_protocol_sequence()

    def update_loop_step_order(self):
        """Goes through all the loop_steps in the sequence, then
        rearranges them in the protocol.

        Parameters
        ----------

        Returns
        -------

        """
        self.general_settings.update_step_config()
        loop_steps = []
        for i in range(self.item_model_sequence.rowCount()):
            item = self.item_model_sequence.item(i, 0)
            sub_steps = treeView_functions.get_substeps(item)
            loop_steps.append((item.data(), sub_steps))
        self.protocol.rearrange_loop_steps(loop_steps)

    def accept(self) -> None:
        """ """
        self.update_loop_step_order()
        self.get_step_config()
        self.check_protocol_name()
        self.check_file_name()
        self.accepted.emit(self.protocol)
        self.is_accepted = True
        self.close()

    def check_file_name(self):
        """check if the filename contains any characters that might cause problems"""
        name = self.general_settings.lineEdit_filename.text()
        invalid_characters = [
            "/",
            "\\",
            ":",
            "*",
            "?",
            '"',
            "<",
            ">",
            "|",
            "+",
            "%",
            "&",
            "#",
            "@",
            "!",
            "$",
            "^",
            "`",
            "~",
        ]
        invals = []
        for char in invalid_characters:
            if char in name:
                invals.append(char)
        if invals:
            raise Exception(
                f"Filename contains invalid characters:\n{' '.join(invals)}"
            )

    def check_protocol_name(self):
        """ """
        name = self.general_settings.lineEdit_protocol_name.text()
        if name in variables_handling.protocols and name != self.old_name:
            raise Exception(f'Protocol name "{name}" already in use!')
        variables_handling.check_variable_name(name, True, self)

    def closeEvent(self, a0: QCloseEvent) -> None:
        """

        Parameters
        ----------
        a0: QCloseEvent :


        Returns
        -------

        """
        name = self.general_settings.lineEdit_protocol_name.text()
        if not self.is_accepted:
            discard_dialog = QMessageBox.question(
                self,
                f"{name} - Discard Changes?",
                f"All changes to {name} will be lost!",
                QMessageBox.Yes | QMessageBox.No,
            )
            if discard_dialog != QMessageBox.Yes:
                a0.ignore()
                return
        a0.accept()
        super().closeEvent(a0)
        self.closing.emit()

    def change_name(self):
        """ """
        self.setWindowTitle(
            f"{self.protocol.name} - Measurement Protocol - NOMAD CAMELS"
        )

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)


if __name__ == "__main__":
    import sys
    from PySide6.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widge = Protocol_Config()
    widge.show()
    sys.exit(app.exec())



import importlib

import importlib_metadata
from PySide6.QtWidgets import (
    QWidget,
    QTableWidgetItem,
    QLabel,
    QMessageBox,
    QTextBrowser,
)
from PySide6.QtCore import Qt

from nomad_camels.gui.instrument_config import Ui_Form

from nomad_camels.frontpanels.instrument_installer import (
    getInstalledDevices,
    Info_Widget,
)

from nomad_camels.utility import variables_handling, device_handling

from nomad_camels.ui_widgets.warn_popup import WarnPopup


class Instrument_Config(Ui_Form, QWidget):
    """ """

    def __init__(self, active_instruments=None, parent=None):
        QWidget.__init__(self, parent=parent)
        self.setupUi(self)
        self.installed_instr, self.packages = getInstalledDevices(return_packages=True)
        active_instruments = active_instruments or {}
        self.active_instruments = {}
        for instrument in active_instruments.values():
            instr = instrument.name
            if instr in self.active_instruments:
                self.active_instruments[instr].append(instrument)
            else:
                self.active_instruments[instr] = [instrument]
        for k in self.installed_instr:
            if k not in self.active_instruments:
                self.active_instruments[k] = []
        for k in self.active_instruments:
            if k not in self.installed_instr:
                WarnPopup(
                    text=f'Instrument type "{k}" in active instruments, but is not installed!',
                    title="instrument not installed",
                )
                # raise Warning(f'Instrument type "{k}" in active instruments, but is not installed!')
        self.tableWidget_instruments.setColumnCount(2)

        self.tableWidget_instruments.verticalHeader().setHidden(True)
        self.current_instr = ""

        self.build_table()
        self.tableWidget_instruments.setMaximumWidth(350)
        self.lineEdit_search.setMaximumWidth(250)
        self.label.setMaximumWidth(100)
        self.lineEdit_search.textChanged.connect(self.build_table)
        self.tableWidget_instruments.clicked.connect(self.table_click)
        self.pushButton_add.clicked.connect(self.add_instance)
        self.pushButton_remove.clicked.connect(self.remove_instance)
        self.info_widge = Info_Widget(self)
        self.info_widge.setHidden(True)
        self.layout().addWidget(self.info_widge, 0, 5, 3, 1)

        self.hide_info = False
        self.toggle_info_hidden()
        self.pushButton_info.clicked.connect(self.toggle_info_hidden)

    def toggle_info_hidden(self):
        if self.hide_info:
            self.pushButton_info.setText("hide info")
        else:
            self.pushButton_info.setText("show info")
        self.hide_info = not self.hide_info
        self.show_hide_info()

    def show_hide_info(self):
        self.info_widge.setHidden(self.hide_info)

    def table_click(self):
        """ """
        self.setCursor(Qt.WaitCursor)
        try:
            ind = self.tableWidget_instruments.selectedIndexes()[0]
            instr = self.tableWidget_instruments.item(ind.row(), 0).text()
            self.get_current_config()
            self.current_instr = instr
            self.label_config.setText(f"Configure: {instr}")
            if instr not in self.packages:
                self.packages[instr] = importlib.import_module(
                    f"nomad_camels_driver_{instr}.{instr}"
                )
            self.config_tabs.clear()
            if not self.active_instruments[instr]:
                self.config_tabs.addTab(
                    QLabel('Add an instrument\ninstance by clicking "+"'),
                    "no instrument",
                )
                self.pushButton_remove.setEnabled(False)
            else:
                self.pushButton_remove.setEnabled(True)
                pack = self.packages[instr]
                for instrument in self.active_instruments[instr]:
                    name = instrument.custom_name
                    inst_widge = pack.subclass_config(
                        parent=self,
                        data=name,
                        settings_dict=instrument.settings,
                        config_dict=instrument.config,
                        additional_info=instrument.additional_info,
                    )
                    self.config_tabs.addTab(inst_widge, name)
                    inst_widge.name_change.connect(self.name_config_changed)
            self.pushButton_add.setEnabled(True)
            self.info_widge.update_texts(instr)
            self.pushButton_info.setHidden(not self.info_widge.info)
            self.show_hide_info()
        finally:
            self.setCursor(Qt.ArrowCursor)

    def name_config_changed(self, new_name):
        """

        Parameters
        ----------
        new_name :


        Returns
        -------

        """
        current_tab = self.config_tabs.currentIndex()
        conf = self.config_tabs.widget(current_tab)
        ind = self.tableWidget_instruments.selectedIndexes()[0]
        instr = self.tableWidget_instruments.item(ind.row(), 0).text()

        if hasattr(conf, "data") and new_name not in self.get_all_names():
            self.active_instruments[instr][current_tab].custom_name = new_name
            conf.data = new_name
            self.config_tabs.setTabText(current_tab, new_name)

    def get_config(self):
        """ """
        self.get_current_config()
        instruments = {}
        for instr in self.active_instruments:
            for instrument in self.active_instruments[instr]:
                instruments[instrument.custom_name] = instrument
        return instruments

    def get_current_config(self):
        """ """
        for i in range(self.config_tabs.count()):
            tab = self.config_tabs.widget(i)
            if not hasattr(tab, "data"):
                continue
            cust_name = self.active_instruments[self.current_instr][i].custom_name
            given_name = tab.lineEdit_custom_name.text()
            if given_name != cust_name:
                WarnPopup(
                    self,
                    f'Instrument name "{given_name}" is either already in use or not allowed (e.g. the instrument`s class is named that way). Using "{cust_name}" instead.',
                    "Instrument name not possible",
                )
            variables_handling.check_variable_name(
                cust_name, parent=self, raise_not_warn=True
            )
            self.active_instruments[self.current_instr][i].settings = tab.get_settings()
            self.active_instruments[self.current_instr][i].config = tab.get_config()
            self.active_instruments[self.current_instr][
                i
            ].additional_info = tab.get_info()
        self.update_channels()

    def update_channels(self):
        """Called when the active devices change.
        The channels in variables_handling are updated with the ones
        provided by the active devices.

        Parameters
        ----------

        Returns
        -------

        """
        variables_handling.channels.clear()
        instruments = {}
        for instr in self.active_instruments:
            for instrument in self.active_instruments[instr]:
                instruments[instrument.custom_name] = instrument
        for key, dev in instruments.items():
            variables_handling.channels.update(dev.get_channels())
            variables_handling.config_channels.update(dev.config_channels)

    def add_instance(self):
        """ """
        self.setCursor(Qt.WaitCursor)
        try:
            ind = self.tableWidget_instruments.selectedIndexes()[0]
            instr = self.tableWidget_instruments.item(ind.row(), 0).text()
            pack = self.packages[instr]
            if not self.active_instruments[instr]:
                self.config_tabs.clear()
                name = instr
                self.pushButton_remove.setEnabled(True)
            else:
                name = instr
                i = 1
                names = self.get_all_names()
                while name in names:
                    name = f"{instr}_{i}"
                    i += 1
            instr_instance = pack.subclass()
            instr_instance.custom_name = name
            self.active_instruments[instr].append(instr_instance)
            single_widge = pack.subclass_config(
                data=name,
                settings_dict=instr_instance.settings,
                config_dict=instr_instance.config,
                additional_info=instr_instance.additional_info,
            )
            self.config_tabs.addTab(single_widge, name)
            single_widge.name_change.connect(self.name_config_changed)
            self.tableWidget_instruments.item(ind.row(), 1).setText(
                str(len(self.active_instruments[instr]))
            )
        finally:
            self.setCursor(Qt.ArrowCursor)

    def get_all_names(self):
        """ """
        names = []
        for instr in self.active_instruments:
            if self.active_instruments[instr]:
                names += [x.custom_name for x in self.active_instruments[instr]]
                names += [self.active_instruments[instr][0].ophyd_class_name]
                names += ["numpy", "np", "StartTime", "ElapsedTime"]
        return names

    def remove_instance(self):
        """ """
        ind = self.config_tabs.currentIndex()
        name = self.config_tabs.tabText(ind)
        remove_dialog = QMessageBox.question(
            self,
            "Remove instrument?",
            f"Are you sure you want to remove the instrument {name}?",
            QMessageBox.Yes | QMessageBox.No,
        )
        if remove_dialog != QMessageBox.Yes:
            return
        self.config_tabs.removeTab(ind)
        instr_ind = self.tableWidget_instruments.selectedIndexes()[0]
        instr = self.tableWidget_instruments.item(instr_ind.row(), 0).text()
        self.active_instruments[instr].pop(ind)
        if not self.active_instruments[instr]:
            self.config_tabs.addTab(
                QLabel('Add an instrument\ninstance by clicking "+"'), "no instrument"
            )
            self.pushButton_remove.setEnabled(False)
        self.tableWidget_instruments.item(instr_ind.row(), 1).setText(
            str(len(self.active_instruments[instr]))
        )

    def build_table(self):
        """ """
        search_text = self.lineEdit_search.text()
        self.installed_instr = getInstalledDevices()
        self.tableWidget_instruments.clear()
        self.tableWidget_instruments.setRowCount(0)
        self.tableWidget_instruments.setHorizontalHeaderLabels(
            ["instrument", "number of instruments"]
        )
        i = 0
        for dev in sorted(self.installed_instr.keys()):
            if dev in device_handling.from_manual_controls:
                continue
            if search_text.lower() not in dev.lower():
                continue
            self.tableWidget_instruments.setRowCount(i + 1)
            item = QTableWidgetItem(dev)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)

            if dev not in self.active_instruments:
                self.active_instruments[dev] = []
            item_n = QTableWidgetItem(str(len(self.active_instruments[dev])))
            item_n.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.tableWidget_instruments.setItem(i, 0, item)
            self.tableWidget_instruments.setItem(i, 1, item_n)
            i += 1
        self.tableWidget_instruments.resizeColumnsToContents()


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widge = Instrument_Config()
    widge.show()
    app.exec()



import sys
import os
from shutil import copyfile

from PySide6.QtWidgets import (
    QDialog,
    QLabel,
    QPushButton,
    QFileDialog,
    QMessageBox,
    QWidget,
    QGridLayout,
    QSpacerItem,
    QSizePolicy,
)
from PySide6.QtGui import QIcon, QCloseEvent

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.utility import variables_handling, fit_variable_renaming
from nomad_camels.tools.driver_builder import Ui_VISA_Device_Builder

from importlib import resources
from nomad_camels import graphics


class Driver_Builder(Ui_VISA_Device_Builder, QDialog):
    """ """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        try:
            self.setWindowIcon(
                QIcon(str(resources.files(graphics) / "camels_icon.png"))
            )
        except:
            pass
        self.setWindowTitle("Driver-builder - NOMAD CAMELS")

        label_custom = QLabel("Custom Channels")
        label_in = QLabel("Read Channels - VISA")
        label_out = QLabel("Set Channels - VISA")
        label_config = QLabel("Config Channels - VISA")
        label_config_in = QLabel("Config Channels - Read Only - VISA")
        label_custom.setStyleSheet("font-weight: bold")
        label_in.setStyleSheet("font-weight: bold")
        label_out.setStyleSheet("font-weight: bold")
        label_config.setStyleSheet("font-weight: bold")
        label_config_in.setStyleSheet("font-weight: bold")
        label_info = QLabel(
            'The values "Write-Format-String", "Return-Parser" and "Query-String" can be left empty, if you want to use a custom function for those!'
        )
        label_info.setStyleSheet("font-weight: bold")
        label_info.setMaximumHeight(20)

        channel_info_config = [
            "Name",
            "Write-Format-String",
            "Input-Type",
            "Return-Parser",
            "Return-Type",
            "Unit",
            "Description",
        ]
        channel_info_in = [
            "Name",
            "Query-String",
            "Return-Parser",
            "Return-Type",
            "Unit",
            "Description",
        ]
        channel_info_out = [
            "Name",
            "Write-Format-String",
            "Return-Parser",
            "Return-Type",
            "Unit",
            "Description",
        ]
        comboboxes_in = {"Return-Type": ["str", "float", "int", "bool"]}
        comboboxes_out = {"Return-Type": ["None", "str", "float", "int", "bool"]}
        comboBoxes_config = dict(comboboxes_out)
        comboBoxes_config.update({"Input-Type": ["str", "float", "bool"]})

        channel_info_custom = [
            "Name",
            "Channel-Type",
            "Input-Type",
            "Unit",
            "Description",
        ]
        comboboxes_custom = {
            "Channel-Type": ["read-only", "set", "config", "config - read-only"],
            "Input-Type": ["str", "float", "bool"],
        }
        self.table_custom = AddRemoveTable(
            headerLabels=channel_info_custom, comboBoxes=comboboxes_custom
        )

        self.table_in = AddRemoveTable(
            headerLabels=channel_info_in, comboBoxes=comboboxes_in
        )
        self.table_out = AddRemoveTable(
            headerLabels=channel_info_out, comboBoxes=comboboxes_out
        )
        self.table_config = AddRemoveTable(
            headerLabels=channel_info_config, comboBoxes=comboBoxes_config
        )
        self.table_config_in = AddRemoveTable(
            headerLabels=channel_info_in, comboBoxes=comboboxes_in
        )

        self.button_build = QPushButton("Build Driver")
        self.button_cancel = QPushButton("Cancel")
        self.button_cancel.clicked.connect(self.close)
        self.button_build.clicked.connect(self.build_driver)
        self.checkBox_VISA.clicked.connect(self.toggle_VISA)
        self.visa_widgets = [
            label_in,
            label_info,
            label_out,
            label_config,
            label_config_in,
            self.table_config,
            self.table_config_in,
            self.table_in,
            self.table_out,
            self.label_read_term,
            self.label_write_term,
            self.label_baud_rate,
            self.lineEdit_baud_rate,
            self.lineEdit_write_term,
            self.lineEdit_read_term,
        ]
        self.toggle_VISA()

        layout = self.layout()

        layout.addWidget(label_custom, 10, 0, 1, 3)
        layout.addWidget(self.table_custom, 11, 0, 4, 3)

        layout.addWidget(label_in, 10, 4, 1, 2)
        layout.addWidget(self.table_in, 11, 4, 1, 2)
        layout.addWidget(label_out, 10, 6, 1, 2)
        layout.addWidget(self.table_out, 11, 6, 1, 2)
        layout.addWidget(label_config_in, 13, 4, 1, 2)
        layout.addWidget(self.table_config_in, 14, 4, 1, 2)
        layout.addWidget(label_config, 13, 6, 1, 2)
        layout.addWidget(self.table_config, 14, 6, 1, 2)
        layout.addWidget(label_info, 12, 4, 1, 3)

        button_widget = QWidget()
        lay = QGridLayout()
        button_widget.setLayout(lay)
        lay.addItem(
            QSpacerItem(
                0,
                0,
                hData=QSizePolicy.Policy.Expanding,
                vData=QSizePolicy.Policy.Minimum,
            ),
            0,
            0,
        )
        lay.addWidget(self.button_build, 0, 1)
        lay.addWidget(self.button_cancel, 0, 2)
        layout.addWidget(button_widget, 50, 0, 1, 8)

        self.adjustSize()

    def toggle_VISA(self):
        visa = self.checkBox_VISA.isChecked()
        for widget in self.visa_widgets:
            widget.setHidden(not visa)
        self.adjustSize()

    def build_driver(self):
        """ """
        dev_name = fit_variable_renaming.replace_name(self.lineEdit_name.text())
        ophyd_name = dev_name.title()
        ophyd_name = fit_variable_renaming.replace_name(ophyd_name)
        search_tags = self.lineEdit_search_tags.text().split()
        visa = self.checkBox_VISA.isChecked()
        configs = inputs = outputs = configs_in = {"Name": []}
        write_term = read_term = ""
        baud_rate = 0
        if visa:
            read_term = fit_variable_renaming.replace_name(
                self.lineEdit_read_term.text()
            )
            write_term = fit_variable_renaming.replace_name(
                self.lineEdit_write_term.text()
            )
            baud_rate = int(self.lineEdit_baud_rate.text())
            inputs = self.table_in.update_table_data()
            outputs = self.table_out.update_table_data()
            configs = self.table_config.update_table_data()
            configs_in = self.table_config_in.update_table_data()

        customs = self.table_custom.update_table_data()
        names = (
            inputs["Name"]
            + outputs["Name"]
            + configs["Name"]
            + configs_in["Name"]
            + customs["Name"]
        )
        for name in names:
            if not variables_handling.check_variable_name(name, parent=self):
                return

        path = variables_handling.device_driver_path or ""
        directory = QFileDialog.getExistingDirectory(
            self, "Select driver's parent directory", f"{path}"
        )
        if not directory:
            return

        sys.path.append(directory)
        if not os.path.isfile(f"{directory}/__init__.py"):
            with open(f"{directory}/__init__.py", "w", encoding="utf-8") as f:
                pass
        module_dir = f"{directory}/{dev_name}"
        driver_dir = f"{module_dir}/nomad_camels_driver_{dev_name}"
        fname = f"{driver_dir}/{dev_name}.py"
        if os.path.isdir(driver_dir) and os.path.isfile(fname):
            answer = QMessageBox.question(
                self,
                "Driver already exists!",
                f"Driver with name {dev_name} already exists.\nDo you want to overwrite the driver?",
                buttons=QMessageBox.Yes | QMessageBox.Cancel,
                defaultButton=QMessageBox.Cancel,
            )
            if answer != QMessageBox.Yes:
                return
        if not os.path.isdir(driver_dir):
            os.makedirs(driver_dir)

        class_string = f"from .{dev_name}_ophyd import {ophyd_name}\n\n"
        class_string += "from nomad_camels.main_classes import device_class\n\n"
        class_string += "class subclass(device_class.Device):\n"
        class_string += "\tdef __init__(self, **kwargs):\n"
        class_string += f'\t\tsuper().__init__(name="{dev_name}", virtual=False, tags={search_tags}, directory="{dev_name}", ophyd_device={ophyd_name}, ophyd_class_name="{ophyd_name}", **kwargs)\n'
        for i, name in enumerate(configs["Name"]):
            if configs["Input-Type"][i] == "str":
                class_string += f'\t\tself.config["{name}"] = ""\n'
            elif configs["Input-Type"][i] == "float":
                class_string += f'\t\tself.config["{name}"] = 0\n'
            elif configs["Input-Type"][i] == "bool":
                class_string += f'\t\tself.config["{name}"] = False\n'
        for i, name in enumerate(customs["Name"]):
            if customs["Channel-Type"][i] == "config":
                if customs["Input-Type"][i] == "str":
                    class_string += f'\t\tself.config["{name}"] = ""\n'
                elif customs["Input-Type"][i] == "float":
                    class_string += f'\t\tself.config["{name}"] = 0\n'
                elif customs["Input-Type"][i] == "bool":
                    class_string += f'\t\tself.config["{name}"] = False\n'

        class_string += "\n\nclass subclass_config(device_class.Simple_Config):\n"
        class_string += '\tdef __init__(self, parent=None, data="", settings_dict=None, config_dict=None, additional_info=None):\n'
        class_string += f'\t\tsuper().__init__(parent, "{dev_name}", data, settings_dict, config_dict, additional_info)\n'
        if visa:
            class_string += '\t\tself.comboBox_connection_type.addItem("Local VISA")\n'
        class_string += "\t\tself.load_settings()\n"

        ophyd_string = "from ophyd import Component as Cpt\n\n"
        if customs:
            ophyd_string += "from nomad_camels.bluesky_handling.custom_function_signal import Custom_Function_Signal, Custom_Function_SignalRO\n"
        if visa:
            ophyd_string += "from nomad_camels.bluesky_handling.visa_signal import VISA_Signal, VISA_Signal_RO, VISA_Device\n\n"
            ophyd_string += f"class {ophyd_name}(VISA_Device):\n"
        else:
            ophyd_string += "from ophyd import Device\n\n"
            ophyd_string += f"class {ophyd_name}(Device):\n"
        open_queries = []
        open_writes = []
        comp_str, opens = make_component_str(inputs, True, False)
        ophyd_string += comp_str
        open_queries += opens
        comp_str, opens = make_component_str(outputs, False, False)
        ophyd_string += comp_str
        open_writes += opens
        comp_str, opens = make_component_str(configs, False, True)
        ophyd_string += comp_str
        open_writes += opens
        comp_str, opens = make_component_str(configs_in, True, True)
        ophyd_string += comp_str
        open_queries += opens

        comp_str, read_names, write_names = make_custom_component(customs)
        ophyd_string += comp_str

        if visa:
            ophyd_string += f'\n\tdef __init__(self, prefix="", *, name, kind=None, read_attrs=None, configuration_attrs=None, parent=None, resource_name="", write_termination="{write_term}", read_termination="{read_term}", baud_rate={baud_rate}, **kwargs):\n'
            ophyd_string += f"\t\tsuper().__init__(prefix=prefix, name=name, kind=kind, read_attrs=read_attrs, configuration_attrs=configuration_attrs, parent=parent, resource_name=resource_name, baud_rate=baud_rate, read_termination=read_termination, write_termination=write_termination, **kwargs)\n"
        else:
            ophyd_string += f'\n\tdef __init__(self, prefix="", *, name, kind=None, read_attrs=None, configuration_attrs=None, parent=None, **kwargs):\n'
            ophyd_string += f"\t\tsuper().__init__(prefix=prefix, name=name, kind=kind, read_attrs=read_attrs, configuration_attrs=configuration_attrs, parent=parent, **kwargs)\n"
        for name in open_queries:
            ophyd_string += f"\t\tself.{name}.query = self.{name}_query_function\n"
        for name in open_writes:
            ophyd_string += f"\t\tself.{name}.write = self.{name}_write_function\n"
        for name in read_names:
            ophyd_string += (
                f"\t\tself.{name}.read_function = self.{name}_read_function\n"
            )
        for name in write_names:
            ophyd_string += f"\t\tself.{name}.put_function = self.{name}_put_function\n"

        ophyd_string += "\n"

        for name in open_queries:
            ophyd_string += f"\tdef {name}_query_function(self):\n"
            ophyd_string += f"\t\tpass\n\n"
        for name in open_writes:
            ophyd_string += f"\tdef {name}_write_function(self, value):\n"
            ophyd_string += f"\t\tpass\n\n"

        for name in read_names:
            ophyd_string += f"\tdef {name}_read_function(self):\n"
            ophyd_string += f"\t\tpass\n\n"
        for name in write_names:
            ophyd_string += f"\tdef {name}_put_function(self, value):\n"
            ophyd_string += f"\t\tpass\n\n"

        readme_string = f"# NOMAD Camels driver for {dev_name}\n\n"
        readme_string += f"Driver for {dev_name} written for the measurement software [NOMAD Camels](https://fau-lap.github.io/NOMAD-CAMELS/).\n\n\n"
        readme_string += "## Documentation\n\n"
        readme_string += "For more information and instruments visit the [documentation](https://fau-lap.github.io/NOMAD-CAMELS/doc/instruments/instruments.html)."

        toml_string = '[build-system]\nrequires = ["setuptools>=61.0"]\nbuild-backend = "setuptools.build_meta"\n\n'
        toml_string += (
            f'[project]\nname = "nomad_camels_driver_{dev_name}"\nversion = "0.1.0"\n'
        )
        toml_string += 'authors = [\n\t{name="enter your name here!", email="contact@email.com"}\n]\n'
        toml_string += (
            f'description = "Instrument driver for {dev_name} for NOMAD CAMELS"\n'
        )
        toml_string += 'readme = "README.md"\n'
        toml_string += 'requires-python = ">=3.9.6"\n'
        toml_string += 'classifiers = [\n\t"Programming Language :: Python :: 3",\n\t"License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",\n]\n\n'
        toml_string += 'dependencies = [\n\t"nomad-camels"'
        if visa:
            toml_string += ',\n\t"pyvisa",\n\t"pyvisa-py"'
        toml_string += "\n]\n\n"
        toml_string += "[project.urls]\n"
        toml_string += (
            '"NOMAD Camels GitHub" = "https://github.com/FAU-LAP/NOMAD-CAMELS"\n'
        )
        toml_string += (
            '"CAMELS Drivers GitHub" = "https://github.com/FAU-LAP/CAMELS_drivers"\n'
        )
        toml_string += '"NOMAD Camels Documentation" = "https://fau-lap.github.io/NOMAD-CAMELS/doc/instruments/instruments.html"\n'

        with open(
            os.path.join(driver_dir, f"{dev_name}.py"), "w", encoding="utf-8"
        ) as f:
            f.write(class_string)
        with open(
            os.path.join(driver_dir, f"{dev_name}_ophyd.py"), "w", encoding="utf-8"
        ) as f:
            f.write(ophyd_string)
        with open(os.path.join(driver_dir, "__init__.py"), "w", encoding="utf-8") as f:
            pass
        with open(os.path.join(module_dir, "README.md"), "w", encoding="utf-8") as f:
            f.write(readme_string)
        with open(
            os.path.join(module_dir, "pyproject.toml"), "w", encoding="utf-8"
        ) as f:
            f.write(toml_string)
        copyfile(
            os.path.join(os.path.dirname(__file__), "../../LICENSE"),
            os.path.join(module_dir, "LICENSE"),
        )

        QMessageBox.information(
            self, "Build finished!", f'The driver "{dev_name}" has been built!'
        )

    def closeEvent(self, a0: QCloseEvent) -> None:
        """

        Parameters
        ----------
        a0: QCloseEvent :


        Returns
        -------

        """
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            a0.ignore()
            return
        super().closeEvent(a0)


def make_custom_component(dic):
    ophyd_string = ""
    read_names = []
    write_names = []
    for i, name in enumerate(dic["Name"]):
        channel_type = dic["Channel-Type"][i]
        read_names.append(name)
        RO = False
        if "read-only" in channel_type:
            RO = True
        else:
            write_names.append(name)
        conf = False
        if "config" in channel_type:
            conf = True
        ophyd_string += f'\t{name} = Cpt(Custom_Function_Signal{"RO" if RO else ""}, name="{name}", '
        if conf:
            ophyd_string += 'kind="config", '
        ophyd_string += f'metadata={{"units": "{dic["Unit"][i]}", "description": "{dic["Description"][i]}"}})\n'
    return ophyd_string, read_names, write_names


def make_component_str(dic, RO, config):
    ophyd_string = ""
    still_open = []
    for i, name in enumerate(dic["Name"]):
        ophyd_string += (
            f'\t{name} = Cpt(VISA_Signal{"_RO" if RO else ""}, name="{name}", '
        )
        if "Write-Format-String" in dic:
            write_str = dic["Write-Format-String"][i]
            if write_str:
                ophyd_string += f'write="{write_str}", '
            else:
                still_open.append(name)
        if "Query-String" in dic:
            query_str = dic["Query-String"][i]
            if query_str:
                ophyd_string += f'query="{query_str}", '
            else:
                still_open.append(name)
        parse_str = dic["Return-Parser"][i]
        if parse_str:
            ophyd_string += f'parse="{parse_str}", '
        return_str = dic["Return-Type"][i]
        if return_str == "None":
            ophyd_string += "parse_return_type=None, "
        else:
            ophyd_string += f'parse_return_type="{return_str}", '
        if config:
            ophyd_string += 'kind="config", '
        ophyd_string += f'metadata={{"units": "{dic["Unit"][i]}", "description": "{dic["Description"][i]}"}})\n'
    return ophyd_string, still_open


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication
    from PySide6.QtCore import QCoreApplication
    from nomad_camels.utility import exception_hook

    sys.excepthook = exception_hook.exception_hook
    app = QCoreApplication.instance()
    if app is None:
        app = QApplication(sys.argv)
    ui = Driver_Builder()
    ui.show()
    app.exec()



import copy

from nomad_camels.main_classes import device_class
from nomad_camels.utility import variables_handling

from .set_panel_device_ophyd import Set_Panel_Device


class subclass(device_class.Device):
    def __init__(self, **kwargs):
        super().__init__(
            name="set_panel_device",
            virtual=True,
            ophyd_device=Set_Panel_Device,
            ophyd_class_name="Set_Panel_Device",
            **kwargs,
        )

    def get_necessary_devices(self):
        channels = self.settings["channel_names"]
        devs = []
        for chan in channels:
            devs.append(variables_handling.channels[chan].device)
        devs = list(set(devs))
        return devs

    def get_channels(self):
        if "group_names" in self.settings:
            channels = copy.deepcopy(super().get_channels())
            n = len(self.settings["group_names"])
            chans = {}
            for i, (nam, val) in enumerate(channels.items()):
                if i >= n:
                    break
                chans[nam] = val
            return chans
        return []


class subclass_config(device_class.Device_Config):
    def __init__(
        self,
        parent=None,
        data="",
        settings_dict=None,
        config_dict=None,
        additional_info=None,
        **kwargs,
    ):
        super().__init__(
            parent,
            "set_panel_device",
            data,
            settings_dict,
            config_dict,
            additional_info,
            **kwargs,
        )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'add_on_manager.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QGridLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QTableWidget,
    QTableWidgetItem,
    QTextEdit,
    QWidget,
)


class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName("Form")
        Form.resize(400, 300)
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.label = QLabel(Form)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.lineEdit_search_name = QLineEdit(Form)
        self.lineEdit_search_name.setObjectName("lineEdit_search_name")
        self.lineEdit_search_name.setMaximumSize(QSize(250, 16777215))

        self.gridLayout.addWidget(self.lineEdit_search_name, 0, 1, 1, 1)

        self.pushButton_install_update_selected = QPushButton(Form)
        self.pushButton_install_update_selected.setObjectName(
            "pushButton_install_update_selected"
        )
        self.pushButton_install_update_selected.setEnabled(True)

        self.gridLayout.addWidget(self.pushButton_install_update_selected, 0, 2, 1, 1)

        self.pushButton_uninstall = QPushButton(Form)
        self.pushButton_uninstall.setObjectName("pushButton_uninstall")
        self.pushButton_uninstall.setEnabled(True)

        self.gridLayout.addWidget(self.pushButton_uninstall, 1, 2, 1, 1)

        self.progressBar = QProgressBar(Form)
        self.progressBar.setObjectName("progressBar")
        self.progressBar.setEnabled(True)
        self.progressBar.setValue(0)
        self.progressBar.setInvertedAppearance(False)

        self.gridLayout.addWidget(self.progressBar, 2, 2, 1, 1)

        self.textEdit_info = QTextEdit(Form)
        self.textEdit_info.setObjectName("textEdit_info")
        self.textEdit_info.setEnabled(True)
        self.textEdit_info.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard | Qt.TextSelectableByMouse
        )

        self.gridLayout.addWidget(self.textEdit_info, 3, 2, 1, 1)

        self.pushButton_info = QPushButton(Form)
        self.pushButton_info.setObjectName("pushButton_info")

        self.gridLayout.addWidget(self.pushButton_info, 4, 2, 1, 1)

        self.extension_table = QTableWidget(Form)
        self.extension_table.setObjectName("extension_table")

        self.gridLayout.addWidget(self.extension_table, 1, 0, 4, 2)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)

    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", "Form", None))
        self.label.setText(QCoreApplication.translate("Form", "Search name:", None))
        self.lineEdit_search_name.setPlaceholderText(
            QCoreApplication.translate("Form", "name", None)
        )
        self.pushButton_install_update_selected.setText(
            QCoreApplication.translate("Form", "Install / Update Selected", None)
        )
        self.pushButton_uninstall.setText(
            QCoreApplication.translate("Form", "Uninstall Selected", None)
        )
        self.pushButton_info.setText(
            QCoreApplication.translate("Form", "hide info", None)
        )

    # retranslateUi



"""This package provides utility for everything regarding devices/instruments
connected to the main UI.

Attributes
----------
    local_packages : dict{"<driver_name>": python-module}
        the loaded modules of local instrument drivers
    running_devices : dict{"<device_name>": ophyd.Device}
        The devices that are already instantiated and currently running. All
        devices are given an attribute `device_run_count` which is increased by
        1 each time a function tries to instantiate a device that is already
        running. Closing the device decreases the run-count by 1. It is only
        really closed, once the run-count reaches 0.
    last_path : str, path
        The path that was last used to search for local drivers. This is used to
        only re-run loading the packages if something changed.
"""

import sys
import os
import importlib
import re

from nomad_camels.utility import variables_handling
from nomad_camels.bluesky_handling import helper_functions

import copy
import pathlib

from PySide6.QtCore import QThread, Signal

local_packages = {}
local_package_paths = {}
running_devices = {}
from_manual_controls = []
last_path = ""


def load_local_packages(tell_local=False):
    """
    Loads the packages of local instrument drivers and returns a dictionary with
    them. If `tell_local`, then the keys will be "local <driver_name>",
    otherwise just "<driver_name>".

    Parameters
    ----------
    tell_local : True
        (Default value = False)

    Returns
    -------
    dict
        Contains the packages with the driver name as keys.
    """
    global local_packages, last_path
    local_instr_path = variables_handling.device_driver_path
    if local_instr_path == last_path:
        if local_packages:
            return local_packages
    else:
        last_path = local_instr_path
        local_packages.clear()
    if not os.path.isdir(local_instr_path):
        return local_packages
    sys.path.append(local_instr_path)
    for f in pathlib.Path(local_instr_path).rglob("*"):
        match = re.match(r"^(nomad[-_]{1}camels[-_]{1}driver[-_]{1})(.*)$", f.name)
        if match:
            try:
                sys.path.append(str(f.parent))
                package = importlib.import_module(f".{match.group(2)}", match.group(0))
                device = package.subclass()
                if tell_local:
                    local_packages[f"local {device.name}"] = package
                else:
                    local_packages[device.name] = package
                local_package_paths[device.name] = str(f.parent)
            except Exception as e:
                print(f, e)
    for f in pathlib.Path("manual_controls").resolve().rglob("*"):
        match = re.match(r"^(nomad[-_]{1}camels[-_]{1}driver[-_]{1})(.*)$", f.name)
        if match:
            try:
                sys.path.append(str(f.parent))
                package = importlib.import_module(f".{match.group(2)}", match.group(0))
                device = package.subclass()
                if tell_local:
                    local_packages[f"local {device.name}"] = package
                else:
                    local_packages[device.name] = package
                local_package_paths[device.name] = str(f.parent)
                from_manual_controls.append(device.name)
            except Exception as e:
                print(f, e)
    return local_packages


def get_channel_from_string(channel):
    """
    Returns the component of the ophyd device that corresponds to the given
    channel. The device has to be instantiated at this point.

    Parameters
    ----------
    channel : str
        The name of the channel (i.e. "<device>.<channel>")

    Returns
    -------
    ophyd.Signal
        The signal / channel found from the string
    """
    dev, chan = channel.split(".")
    if dev not in running_devices:
        raise Exception(f"Device {dev} is needed, but not yet instantiated!")
    device = running_devices[dev]
    return getattr(device, chan)


def get_funtion_from_string(func_name):
    dev, func = func_name.split(".")
    if dev not in running_devices:
        raise Exception(f"Device {dev} is needed, but not yet instantiated!")
    device = running_devices[dev]
    return getattr(device, func)


def get_channels_from_string_list(channel_list, as_dict=False):
    """
    Goes through the given channel_list and if they are valid channels in
    CAMELS, their ophyd representation is called by `get_channel_from_string`.

    Parameters
    ----------
    channel_list : list[str]
        List of the channels in CAMELS-representation
        (i.e. "<device_name>_<channel_name>")
    as_dict : bool
        (Default value = False)
        if True, the returned channels will be a dictionary with the original
        list serving as keys

    Returns
    -------
    list[ophyd.Signal], dict
        A list of the ophyd representations of `channel_list`.
        If `as_dict` is True, it is a dictionary with the shape
        {'channel_name': ophyd.Signal}
    """
    if as_dict:
        channels = {}
    else:
        channels = []
    for channel in channel_list:
        chan = channel
        if chan == "None":
            channels.append(None)
            continue
        if channel in variables_handling.channels:
            chan = variables_handling.channels[channel]
        if as_dict:
            channels[channel] = get_channel_from_string(chan.name)
        else:
            channels.append(get_channel_from_string(chan.name))
    return channels


def get_functions_from_string_list(func_list):
    funcs = []
    for func in func_list:
        if func == "None":
            funcs.append(None)
            continue
        funcs.append(get_funtion_from_string(func))
    return funcs


def start_devices_from_channel_list(channel_list, skip_config=False):
    """
    Instantiates the ophyd devices that are needed by the given channels.
    Returns the ophyd devices and their metadata.

    Parameters
    ----------
    channel_list : list[str]
        List of the channels, for which the devices should be started.

    Returns
    -------
    devs : dict{"<device_name>": ophyd.Device}
        Dictionary of the started devices (or if they were already running,
        their currently running instance)
    dev_data : dict{"<device_name>": dict}
        Dictionary of the devices' metadata (i.e. config, settings...)
    """
    dev_list = set()
    for channel in channel_list:
        dev_list.add(variables_handling.channels[channel].device)
    dev_list = list(dev_list)
    devs, dev_data = instantiate_devices(dev_list, skip_config=skip_config)
    return devs, dev_data


def instantiate_devices(device_list, skip_config=False):
    """
    Starts the given devices, or increases their run-count by 1 and returns
    them together with their settings.

    Parameters
    ----------
    device_list : list[str]
        The devices (as they are named in CAMELS) that should be started

    Returns
    -------
    devices : dict{"<device_name>": ophyd.Device}
        Dictionary of the started devices (or if they were already running,
        their currently running instance)
    device_config : dict{"<device_name>": dict}
        Dictionary of the devices' metadata (i.e. config, settings...)
    """
    device_config = {}
    devices = {}
    started_devs = []
    try:
        for dev in device_list:
            # getting all the settings
            device = variables_handling.devices[dev]
            classname = device.ophyd_class_name
            if skip_config:
                config = {}
            else:
                config = copy.deepcopy(device.get_config())
            settings = copy.deepcopy(device.get_settings())
            additional_info = copy.deepcopy(device.get_additional_info())
            if "connection" in settings:
                conn = settings.pop("connection")
                if "type" in conn:
                    conn.pop("type")
                settings.update(conn)
            if "idn" in settings:
                settings.pop("idn")
            extra_settings = {}
            non_strings = []
            for key in settings:
                if key.startswith("!non_string!_"):
                    extra_settings[key.replace("!non_string!_", "")] = settings[key]
                    non_strings.append(key)
            for s in non_strings:
                settings.pop(s)
            additional_info["device_class_name"] = classname
            extra_settings.update(settings)

            extra_config = {}
            non_strings = []
            for key in config:
                if key.startswith("!non_string!_"):
                    extra_config[key.replace("!non_string!_", "")] = config[key]
                    non_strings.append(key)
            for s in non_strings:
                config.pop(s)
            config.update(extra_config)

            # instantiating ophyd-device
            print(f"connecting {dev}")
            if dev in running_devices:
                ophyd_device = running_devices[dev]
                ophyd_device.device_run_count += 1
            else:
                ophyd_device = device.ophyd_class(f"{dev}:", name=dev, **extra_settings)
                ophyd_device.device_run_count = 1
                running_devices[dev] = ophyd_device
            ophyd_device.wait_for_connection()
            configs = ophyd_device.configure(config)[1]
            devices[dev] = ophyd_device

            # updating the config data
            device_config[dev] = {"settings": {}}
            device_config[dev]["settings"].update(
                helper_functions.simplify_configs_dict(configs)
            )
            device_config[dev]["settings"].update(settings)
            device_config[dev].update(additional_info)
            device_config[dev]["instrument_camels_channels"] = {
                name: dict(data.__dict__) for name, data in device.channels.items()
            }  # dict() is necessary to make a copy of the data (dict is a mutable object)
            for watchdog in variables_handling.watchdogs.values():
                if dev in watchdog.get_device_list():
                    watchdog.add_device(dev, ophyd_device)
            started_devs.append(dev)
    except Exception as e:
        close_devices(started_devs)
        raise Exception(e)
    return devices, device_config


class InstantiateDevicesThread(QThread):
    """
    Thread for starting devices in the background.
    """

    exception_raised = Signal(Exception)
    successful = Signal()

    def __init__(self, device_list, channels=False, skip_config=False):
        super().__init__()
        self.channels = channels
        self.device_list = device_list
        self.skip_config = skip_config
        main_thread_devs = []
        if channels:
            for channel in device_list:
                if channel not in variables_handling.channels:
                    raise Warning(
                        f"Trying to use channel {channel}, but it is not defined!"
                    )
                chan = variables_handling.channels[channel]
                if chan.device not in variables_handling.devices:
                    raise Warning(
                        f"Trying to use channel {channel}, but the corresponding device {chan.device} is not defined!"
                    )
                dev = variables_handling.devices[chan.device]
                if dev.main_thread_only:
                    main_thread_devs.append(channel)
        else:
            for device in device_list:
                if device not in variables_handling.devices:
                    raise Warning(
                        f"Trying to start device {device}, but it is not even defined!"
                    )
                dev = variables_handling.devices[device]
                if dev.main_thread_only:
                    main_thread_devs.append(device)
                    for d in dev.get_necessary_devices():
                        if d not in main_thread_devs:
                            main_thread_devs.insert(0, d)
        for dev in main_thread_devs:
            self.device_list.remove(dev)
        if self.channels:
            self.devices, self.device_config = start_devices_from_channel_list(
                main_thread_devs, skip_config=skip_config
            )
        else:
            self.devices, self.device_config = instantiate_devices(
                main_thread_devs, skip_config=skip_config
            )
        # uncomment next line for testing purposes
        # self.run()

    def run(self):
        try:
            if self.channels:
                devices, device_config = start_devices_from_channel_list(
                    self.device_list, skip_config=self.skip_config
                )
            else:
                devices, device_config = instantiate_devices(
                    self.device_list, skip_config=self.skip_config
                )
            self.devices.update(devices)
            self.device_config.update(device_config)
            self.successful.emit()
        except Exception as e:
            self.exception_raised.emit(e)


def close_devices(device_list):
    """
    Closes the given devices (calling their `finalize_steps`), or decreases
    their run-count by 1.

    Parameters
    ----------
    device_list : list[str]
        The devices (as they are named in CAMELS) that should be closed
    """
    dependent_devices = []
    for dev in device_list:
        device = variables_handling.devices[dev]
        dependent_devices += device.get_necessary_devices()
    dependent_devices = list(set(dependent_devices))
    for dev in reversed(device_list):
        if dev not in running_devices:
            continue
            raise Warning(f"Trying to close device {dev}, but it is not even running!")
        if dev in dependent_devices:
            continue
        close_single_device(dev)
    for dev in dependent_devices:
        if dev not in running_devices:
            continue
        close_single_device(dev)


def close_single_device(dev):
    ophyd_dev = running_devices[dev]
    ophyd_dev.device_run_count -= 1
    if ophyd_dev.device_run_count == 0:
        running_devices.pop(dev)
        if hasattr(ophyd_dev, "finalize_steps") and callable(ophyd_dev.finalize_steps):
            ophyd_dev.finalize_steps()
        for watchdog in variables_handling.watchdogs.values():
            if dev in watchdog.get_device_list():
                watchdog.remove_device(dev, ophyd_dev)



"""This module is used to change the UI-theme of CAMELS. It provides some
default color palettes `light_palette` and `dark_palette`."""

from PySide6.QtWidgets import QApplication, QMainWindow, QStyleFactory
from PySide6.QtGui import QPalette, QColor, QColorConstants
from qt_material import apply_stylesheet


light_palette = QPalette(QColor(225, 225, 225), QColor(238, 238, 238))
light_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))

dark_palette = QPalette()
dark_palette.setColor(QPalette.Window, QColor(53, 53, 53))
dark_palette.setColor(QPalette.WindowText, QColorConstants.White)
dark_palette.setColor(QPalette.Base, QColor(25, 25, 25))
dark_palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
dark_palette.setColor(QPalette.ToolTipBase, QColorConstants.White)
dark_palette.setColor(QPalette.ToolTipText, QColorConstants.White)
dark_palette.setColor(QPalette.Text, QColorConstants.White)
dark_palette.setColor(QPalette.Button, QColor(53, 53, 53))
dark_palette.setColor(QPalette.ButtonText, QColorConstants.White)
dark_palette.setColor(QPalette.BrightText, QColorConstants.Red)
dark_palette.setColor(QPalette.Link, QColor(42, 130, 218))
dark_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
dark_palette.setColor(QPalette.HighlightedText, QColorConstants.Black)


def change_theme(theme, main_app=None, material_theme=None, dark_mode=False):
    """

    Parameters
    ----------
    theme : str
        The name of the used theme. Possible values are the themes from
        QStyleFactory and "qt-material".
    main_app : QMainWindow
        (Default value = None)
        The main UI-window. If it is None, it is looked for with
        `QApplication.instance()`.
    material_theme : str
        (Default value = None)
        If `theme=="qt-material"`, the qt-material color palette is chosen with
        this theme name.
    dark_mode : bool
        (Default value = False)
        If True, the color palettes are switched to dark mode.
    """
    if main_app is None:
        main_app = QApplication.instance()
        if main_app is None:
            raise RuntimeError("MainApp not found for changing the color theme.")
    if theme in QStyleFactory.keys():
        main_app.setStyleSheet("")
        palette = None
        if dark_mode:
            palette = dark_palette
        else:
            palette = light_palette
        if theme == "windowsvista":
            main_app.setStyle(QStyleFactory.create("windowsvista"))
        main_app.setPalette(palette)
        main_app.setStyle(QStyleFactory.create(theme))
        if palette:
            main_app.setPalette(palette)
    elif theme == "qt-material":
        if not isinstance(material_theme, str):
            return
        if dark_mode:
            if material_theme.endswith("500"):
                material_theme = material_theme[:-4]
            xml = f"dark_{material_theme}.xml"
        else:
            xml = f"light_{material_theme}.xml"
        apply_stylesheet(main_app, theme=xml)



from ophyd import EpicsSignalRO
from ophyd.status import Status


class TriggerEpicsSignalRO(EpicsSignalRO):
    """Subclass that implements a simple 'timed' reading when called via
    bluesky's 'trigger_and_read'.
    Note that each PV that shall be read via this class also needs a
    second PV with the same name+:trig. That PV should provide a FLNK to
    the original PV.

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(
        self, read_pv, *, timeout=10, string=False, name=None, no_mdel=False, **kwargs
    ):
        super().__init__(
            read_pv,
            string=string,
            name=name,
            timeout=timeout,
            auto_monitor=False,
            **kwargs,
        )
        self.stat = None
        self.trigger_pv = self.cl.get_pv(f"{read_pv}:trig")
        self.subscribe(self.callback_method)
        self.last_time = self.timestamp
        self.no_mdel = no_mdel
        self.triggering = True

    def callback_method(self, **kwargs):
        """If there is a status object from the trigger-method, it will
        be set to finished.

        Parameters
        ----------
        **kwargs :


        Returns
        -------

        """
        if self.stat is not None and not self.no_mdel and self.triggering:
            self.stat.set_finished()
            # self.stat = None

    def trigger(self):
        """Returns a status object that will be set to finished, when
        the PV-value is updated. Sets the trigger-PV to 1, thus
        triggering the process of the original PV.

        Parameters
        ----------

        Returns
        -------

        """
        self.stat = Status(self, timeout=self.timeout)
        if self.triggering:
            self.trigger_pv.put(1)
        if self.no_mdel or not self.triggering:
            self.stat.set_finished()
        return self.stat



from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.utility import variables_handling


class Set_Variables(Loop_Step):
    """
    This step enables setting variables to a different value during the protocol.

    Attributes
    ----------
    variables_values : dict{'Variable': list[str], 'Value': list[str]}
        Contains a list of the variables and a list of the respective values to
        which they should be set.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Set Variables"
        if step_info is None:
            step_info = {}
        self.variables_values = (
            step_info["variables_values"]
            if "variables_values" in step_info
            else {"Variable": [], "Value": []}
        )

    def get_protocol_string(self, n_tabs=1):
        """Evaluates the values for the variables, then updates the namespace."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        for i, variable in enumerate(self.variables_values["Variable"]):
            val = variables_handling.get_write_from_data_type(
                self.variables_values["Value"][i]
            )
            if isinstance(val, str):
                protocol_string += f"{tabs}{variable} = eva.eval({val})\n"
            else:
                protocol_string += f"{tabs}{variable} = {val}\n"
            protocol_string += f'{tabs}namespace["{variable}"] = {variable}\n'
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Shows the variables and values."""
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]} - {self.variables_values}\n"
        return short_string


class Set_Variables_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Set_Variables, parent=None):
        super().__init__(parent, loop_step)
        self.variables_table = AddRemoveTable(
            tableData=loop_step.variables_values, checkstrings=[1]
        )
        self.layout().addWidget(self.variables_table, 1, 0, 1, 5)

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.loop_step.variables_values = self.variables_table.update_table_data()



from PySide6.QtWidgets import QWidget, QLabel, QCheckBox, QGridLayout, QMessageBox
from PySide6.QtCore import Signal
from PySide6.QtGui import QStandardItemModel

from nomad_camels.main_classes.loop_step import Loop_Step_Container, Loop_Step_Config
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable


class If_Loop_Step(Loop_Step_Container):
    """A loopstep providing an if-case selection with a variable number
    of `elif`.

    Attributes
    ----------
    condition : str
        The condition which is used for the if-conditional (should be valid
        python code)
    use_else : bool
        whether to add an `else`-case
    elifs : list of str
        string-list of the conditions for a variable number of `elif` cases
    """

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "If"
        self.has_children = False
        if step_info is None:
            step_info = {}
        self.condition = step_info["condition"] if "condition" in step_info else "!"
        self.use_else = step_info["use_else"] if "use_else" in step_info else False
        self.elifs = step_info["elifs"] if "elifs" in step_info else []
        self.end_protocol = step_info["end_protocol"] if "end_protocol" in step_info else False
        self.update_children()

    def append_to_model(self, item_model: QStandardItemModel, parent=None):
        """Overwritten, so that nothing can be dropped into the main step."""
        item = super(If_Loop_Step, self).append_to_model(item_model, parent)
        item.setDropEnabled(False)
        return item

    def update_children(self):
        """Updates the if-substeps provided by this step, giving them
        the names corresponding to the conditions."""
        if not self.children:
            self.children.append(If_Sub_Step(f"{self.name}_{self.condition}"))
        n_children = 1 + len(self.elifs) + (1 if self.use_else else 0)
        if self.use_else and not isinstance(self.children[-1], Else_Sub_Step):
            self.children.append(Else_Sub_Step(f"{self.name}_Else"))
        elif not self.use_else and isinstance(self.children[-1], Else_Sub_Step):
            self.children.pop()
        pops = []
        for i, child in enumerate(self.children):
            if i == 0:
                child.name = f"{self.name}_{self.condition}"
            elif self.use_else and i == len(self.children) - 1:
                child.name = f"{self.name}_Else"
            else:
                try:
                    child.name = f"{self.name}_{self.elifs[i-1]}"
                except IndexError:
                    pops.append(i)
        for i in reversed(pops):
            self.children.pop(i)
        diff = n_children - len(self.children)
        if diff > 0:
            for cond in self.elifs[-diff:]:
                if self.use_else:
                    self.children.insert(-1, Elif_Sub_Step(f"{self.name}_{cond}"))
                else:
                    self.children.append(Elif_Sub_Step(f"{self.name}_{cond}"))
        elif diff < 0:
            raise Exception("something went wrong with if-children!")

    def get_protocol_string(self, n_tabs=1):
        """Putting together the children of all if-substeps and the conditions"""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f'{tabs}if eva.eval("{self.condition}"):\n'
        protocol_string += self.children[0].get_children_strings(n_tabs + 1)
        if self.end_protocol:
            protocol_string += f"{tabs}\treturn\n"
            return protocol_string
        for i, el in enumerate(self.elifs):
            child = self.children[i + 1]
            protocol_string += f'{tabs}elif eva.eval("{el}"):\n'
            protocol_string += child.get_children_strings(n_tabs + 1)
        if self.use_else:
            protocol_string += f"{tabs}else:\n"
            protocol_string += self.children[-1].get_children_strings(n_tabs + 1)
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Shows all the conditions and the steps' children"""
        tabs = "\t" * n_tabs
        short_string = f"{tabs}if {self.condition}:\n"
        for child in self.children[0].children:
            short_string += child.get_protocol_short_string(n_tabs + 1)
        for i, el in enumerate(self.elifs):
            short_string += f"{tabs}else if {el}:\n"
            for child in self.children[i + 1].children:
                short_string += child.get_protocol_short_string(n_tabs + 1)
        if self.use_else:
            short_string += f"{tabs}else:\n"
            for child in self.children[-1].children:
                short_string += child.get_protocol_short_string(n_tabs + 1)
        return short_string


class If_Sub_Step(Loop_Step_Container):
    """Simple sub-step that represents the if-condition."""

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "If_Sub"

    def append_to_model(self, item_model: QStandardItemModel, parent=None):
        """Overwritten, so that nothing can be dropped into the main step. The
        Sub-Step cannot be dragged out from the main step."""
        item = super().append_to_model(item_model, parent)
        item.setDragEnabled(False)
        item.setEnabled(False)
        return item


class Elif_Sub_Step(If_Sub_Step):
    """Simple sub-step that represents an elif-condition."""

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "Elif_Sub"


class Else_Sub_Step(If_Sub_Step):
    """Simple sub-step that represents the else-condition."""

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "Else_Sub"


class Sub_Step_Config(Loop_Step_Config):
    """Configuration for the substeps, disabling the name-widget."""

    def __init__(self, loop_step: Loop_Step_Container, parent=None):
        super().__init__(parent, loop_step)
        self.name_widget.setEnabled(False)


class If_Step_Config(Loop_Step_Config):
    """Configuration-Widget for the if-loop step."""

    def __init__(self, loop_step: If_Loop_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = If_Step_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)
        self.sub_widget.update_config.connect(self.update_step_config)

    def update_step_config(self):
        """Overwritten, so that the loopstep updates its substeps for
        correctly building the protocol sequence.

        Parameters
        ----------

        Returns
        -------

        """
        self.sub_widget.update_condition(True)
        self.loop_step.update_children()
        super().update_step_config()


class If_Step_Config_Sub(QWidget):
    """This widget consists of a line for the if-condition an
    AddRemoveTable for the elif-conditions and a checkbox for whether to
    use an else-case.

    Parameters
    ----------

    Returns
    -------

    """

    update_config = Signal()

    def __init__(self, loop_step: If_Loop_Step, parent=None):
        super().__init__(parent)
        self.loop_step = loop_step

        label = QLabel("Condition:")
        self.lineEdit_condition = Variable_Box(self)
        self.lineEdit_condition.setText(loop_step.condition)
        self.lineEdit_condition.textChanged.connect(self.update_condition)

        self.elif_table = AddRemoveTable(
            parent=self,
            headerLabels=[],
            tableData=self.loop_step.elifs,
            title="Elif-cases:",
            checkstrings=0,
            askdelete=True,
        )
        self.elif_table.sizechange.connect(self.update_condition)

        self.checkBox_use_else = QCheckBox("Use Else")
        self.checkBox_use_else.setChecked(self.loop_step.use_else)
        self.checkBox_use_else.clicked.connect(self.else_change)

        # Widget to end protocol if the condition is met
        self.end_protocol_checkbox = QCheckBox("End protocol if condition is true")
        self.end_protocol_checkbox.setChecked(self.loop_step.end_protocol)
        self.end_protocol_checkbox.clicked.connect(self.update_end_protocol_checkbox)


        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.lineEdit_condition, 0, 1)
        layout.addWidget(self.checkBox_use_else, 1, 0, 1, 2)
        layout.addWidget(self.end_protocol_checkbox, 2, 0, 1, 2)
        layout.addWidget(self.elif_table, 10, 0, 1, 2)
        self.setLayout(layout)

    def else_change(self):
        """When removing the else-case, first ask the user if they are
        sure, since substeps will be deleted as well.

        Parameters
        ----------

        Returns
        -------

        """
        if self.loop_step.use_else:
            dialog = QMessageBox.question(
                self,
                "Remove Else case?",
                "Removing the Else-case also deletes its children",
                QMessageBox.Yes | QMessageBox.No,
            )
            if dialog != QMessageBox.Yes:
                self.checkBox_use_else.setChecked(True)
                return
        self.update_condition()
    
    def update_end_protocol_checkbox(self):
        """Update the end protocol checkbox"""
        self.loop_step.end_protocol = self.end_protocol_checkbox.isChecked()
    
    def update_condition(self, called=False):
        """Updating the loopstep. The signal is not emitted if the
        function is called from outside, so there is no infinite loop.

        Parameters
        ----------
        called :
             (Default value = False)

        Returns
        -------

        """
        self.loop_step.use_else = self.checkBox_use_else.isChecked()
        self.loop_step.elifs = self.elif_table.update_table_data()
        self.loop_step.condition = self.lineEdit_condition.text()
        if not called:
            self.update_config.emit()



from PySide6.QtWidgets import QCheckBox

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table

import ast


class Set_Value_Popup(Loop_Step):
    """
    This step provides the possibility for the user to set a channel or variable
    to a specific value at runtime with a popup. The protocol execution is
    paused until the user is finished.

    Attributes
    ----------
    variables : list[str]
        A list of the variables that should be set by the user.
    channels : list[str]
        A list of the channels that should be set by the user.
    free_variables : bool
        If True, the user has the possibility to set any variable freely.
    free_channels : bool
        If True, the user has the possibility to set any channel freely.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Set Value Popup"
        if step_info is None:
            step_info = {}
        self.variables = step_info["variables"] if "variables" in step_info else []
        self.combo_list = step_info["combo_list"] if "combo_list" in step_info else []
        self.channels = step_info["channels"] if "channels" in step_info else []
        self.free_variables = (
            step_info["free_variables"] if "free_variables" in step_info else False
        )
        self.free_channels = (
            step_info["free_channels"] if "free_channels" in step_info else False
        )

    def get_protocol_string(self, n_tabs=1):
        """The value popup box is executed. The protocol waits until its
        completion. Then all the variables and then the channels are set to the
        given values."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f'{tabs}boxes["values_{self.name}"].done_flag = False\n'
        protocol_string += f'{tabs}boxes["values_{self.name}"].helper.executor.emit()\n'
        protocol_string += f'{tabs}while not boxes["values_{self.name}"].done_flag:\n'
        protocol_string += f"{tabs}\tyield from bps.sleep(0.1)\n"
        protocol_string += f'{tabs}if boxes["values_{self.name}"].was_accepted:\n'
        protocol_string += f'{tabs}\tfor var, val in boxes["values_{self.name}"].set_variables.items():\n'
        protocol_string += f"{tabs}\t\tnamespace[var] = eva.eval(str(val))\n"
        protocol_string += f'{tabs}\tfor chan, val in boxes["values_{self.name}"].set_channels.items():\n'
        protocol_string += f'{tabs}\t\tdev, channel_name = boxes["values_{self.name}"].channel_devs[chan]\n'
        protocol_string += f'{tabs}\t\tyield from bps.abs_set(getattr(devs[dev], channel_name), eva.eval(str(val)), group="A")\n'
        protocol_string += f'{tabs}\tyield from bps.wait("A")\n'
        return protocol_string

    def get_add_main_string(self):
        """Adds the setup of the box."""
        add_main_string = super().get_add_main_string()
        add_main_string += f'\tboxes["values_{self.name}"] = helper_functions.Value_Box("", "Set Values!", {self.variables}, {self.channels}, {self.free_variables}, {self.free_channels}, devs=devs, comboboxes={self.combo_list})\n'
        return add_main_string


class Set_Value_Popup_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Set_Value_Popup, parent=None):
        super().__init__(parent, loop_step)
        combo_list = loop_step.combo_list or ['' for _ in loop_step.variables]
        table_data = {'variable': loop_step.variables,
                      'combobox-list': combo_list}
        self.variables_table = AddRemoveTable(
            tableData=table_data, title="Variables", headerLabels=['variable', 'combobox-list']
        )
        self.variables_table.setToolTip("If you want to allow free setting, keep combobox-list empty, otherwise you can fill it with a list of possible values.\nExample: ['value1', 'value2'] or [2, 3, 4]")
        labels = ["use", "channel"]
        info_dict = {"channel": loop_step.channels}
        self.channels_table = Channels_Check_Table(
            self, labels, info_dict=info_dict, title="Channels", only_output=True
        )
        self.checkBox_free_variables = QCheckBox("Allow free variable setting")
        self.checkBox_free_channels = QCheckBox("Allow free channel setting")

        self.layout().addWidget(self.checkBox_free_channels, 1, 0, 1, 5)
        self.layout().addWidget(self.checkBox_free_variables, 2, 0, 1, 5)
        self.layout().addWidget(self.channels_table, 3, 0, 1, 5)
        self.layout().addWidget(self.variables_table, 4, 0, 1, 5)

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.loop_step.variables = self.variables_table.update_table_data()['variable']
        self.loop_step.combo_list = [ast.literal_eval(x) if x else "" for x in self.variables_table.update_table_data()['combobox-list']]
        self.loop_step.channels = self.channels_table.get_info()["channel"]
        self.loop_step.free_channels = self.checkBox_free_channels.isChecked()
        self.loop_step.free_variables = self.checkBox_free_variables.isChecked()



from ophyd import Component as Cpt
from ophyd import Device

from nomad_camels.bluesky_handling.custom_function_signal import Custom_Function_Signal
from nomad_camels.utility import device_handling

import time as ttime


class Set_Panel_Device(Device):
    group_1 = Cpt(Custom_Function_Signal, name="group_1")
    group_2 = Cpt(Custom_Function_Signal, name="group_2")
    group_3 = Cpt(Custom_Function_Signal, name="group_3")
    group_4 = Cpt(Custom_Function_Signal, name="group_4")
    group_5 = Cpt(Custom_Function_Signal, name="group_5")
    group_6 = Cpt(Custom_Function_Signal, name="group_6")
    group_7 = Cpt(Custom_Function_Signal, name="group_7")
    group_8 = Cpt(Custom_Function_Signal, name="group_8")
    group_9 = Cpt(Custom_Function_Signal, name="group_9")
    group_10 = Cpt(Custom_Function_Signal, name="group_10")
    group_11 = Cpt(Custom_Function_Signal, name="group_11")
    group_12 = Cpt(Custom_Function_Signal, name="group_12")
    group_13 = Cpt(Custom_Function_Signal, name="group_13")
    group_14 = Cpt(Custom_Function_Signal, name="group_14")
    group_15 = Cpt(Custom_Function_Signal, name="group_15")
    group_16 = Cpt(Custom_Function_Signal, name="group_16")
    group_17 = Cpt(Custom_Function_Signal, name="group_17")
    group_18 = Cpt(Custom_Function_Signal, name="group_18")
    group_19 = Cpt(Custom_Function_Signal, name="group_19")
    group_20 = Cpt(Custom_Function_Signal, name="group_20")

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        set_vals=None,
        group_names=None,
        channels=None,
        **kwargs,
    ):
        if "channel_names" in kwargs:
            kwargs.pop("channel_names")
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.set_vals = set_vals or {}
        self.group_names = group_names or []
        self.channels = channels or {}
        self.order = [
            self.group_1,
            self.group_2,
            self.group_3,
            self.group_4,
            self.group_5,
            self.group_6,
            self.group_7,
            self.group_8,
            self.group_9,
            self.group_10,
            self.group_11,
            self.group_12,
            self.group_13,
            self.group_14,
            self.group_15,
            self.group_16,
            self.group_17,
            self.group_18,
            self.group_19,
            self.group_20,
        ]
        for i, group in enumerate(self.order):
            group.put_function = lambda x, n=i: self.set_value(x, n)

        comps = list(self.component_names)
        for i, comp in enumerate(self.component_names):
            if i <= len(self.group_names):
                continue
            comps.remove(comp)
        self.component_names = tuple(comps)

    def set_value(self, i, n):
        for j, channel in enumerate(self.set_vals[n][i]["channel"]):
            value = self.set_vals[n][i]["value"][j]
            try:
                value = int(value)
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    try:
                        value = bool(value)
                    except ValueError:
                        pass
            chan = self.channels[channel]
            if isinstance(chan, str):
                chan = device_handling.get_channel_from_string(chan)
            chan.put(value)



active_eln_context = None


class ELN_Context:
    def __init__(self, ui):
        self.comboBox_user_type = ui.comboBox_user_type
        self.user_widget = ui.user_widget
        self.sample_widget = ui.sample_widget
        self.session_upload_widget = ui.session_upload_widget
        self.comboBox_user_type = ui.comboBox_user_type
        self.run_done_file_signal = ui.run_done_file_signal
        self.extension_user = ui.extension_user
        self.extension_sample = ui.extension_sample
        self.selection_function = None



from PySide6.QtWidgets import (
    QWidget,
    QTableWidgetItem,
    QLabel,
    QTableWidget,
    QGridLayout,
    QLineEdit,
    QMenu,
    QVBoxLayout,
    QWidgetAction,
)
from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QFont

from nomad_camels.utility import variables_handling


class Channels_Check_Table(QWidget):
    """ """

    def __init__(
        self,
        parent,
        headerLabels=None,
        only_output=False,
        info_dict=None,
        checkstrings=None,
        title="",
        channels=None,
        use_configs=False,
    ):
        super().__init__(parent)
        if use_configs:
            self.channels = channels or variables_handling.config_channels
        else:
            self.channels = channels or variables_handling.channels
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)
        self.only_output = only_output
        self.headerLabels = headerLabels or []
        self.checkstrings = checkstrings or []
        self.info_dict = info_dict or {}
        if "channel" not in self.info_dict:
            self.info_dict["channel"] = []
        for lab in self.headerLabels[2:]:
            if lab not in self.info_dict:
                self.info_dict[lab] = []

        layout = QGridLayout()
        self.tableWidget_channels = QTableWidget()
        self.tableWidget_channels.setHorizontalHeaderLabels(self.headerLabels)
        label_search = QLabel("Search:")
        self.tableWidget_channels.clicked.connect(
            self.tableWidget_channels.resizeColumnsToContents
        )
        self.lineEdit_search = QLineEdit()
        # Add default text to the search bar
        self.lineEdit_search.setPlaceholderText("Filter instruments ...")
        self.lineEdit_search.textChanged.connect(self.change_search)

        self.setLayout(layout)
        if title:
            title_label = QLabel(title)
            layout.addWidget(title_label, 0, 0, 1, 2)
            font = QFont()
            font.setBold(True)
            title_label.setStyleSheet("font-size: 9pt")
            title_label.setFont(font)
        layout.addWidget(label_search, 1, 0)
        layout.addWidget(self.lineEdit_search, 1, 1)
        layout.addWidget(self.tableWidget_channels, 2, 0, 1, 2)
        layout.setContentsMargins(0, 0, 0, 0)
        self.build_channels_table()
        self.tableWidget_channels.itemChanged.connect(self.check_string)
        self.tableWidget_channels.clicked.connect(self.check_change)

    def context_menu(self, pos):
        """Generates the right-click-menu.
        There are entries for inserting (replace) and appending the
        variables, channels, functions and operators.

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        if self.parent() and hasattr(self.parent(), "loop_step"):
            prot = self.parent().loop_step.protocol
            variables_handling.protocol_variables = prot.variables
            prot.update_variables()
            variables_handling.loop_step_variables = prot.loop_step_variables
        self.menu = QMenu()
        search_bar = QLineEdit(self.menu)
        search_bar.setPlaceholderText("Search...")
        # Setting up layout for the menu
        layout = QVBoxLayout()
        layout.addWidget(search_bar)
        container = QWidget()
        container.setLayout(layout)

        # Adding the container widget to the menu
        action = QWidgetAction(self.menu)
        action.setDefaultWidget(container)
        self.menu.addAction(action)

        # putting the returned actions somewhere is necessary, otherwise
        # there will be none inside the single menus
        (self.channel_menu, self.variable_menu, self.function_menu), _ = (
            variables_handling.get_menus(self.insert_variable)
        )
        self.menu.addMenu(self.channel_menu)
        self.menu.addMenu(self.variable_menu)
        self.menu.addMenu(self.function_menu)
        # menu.addMenu(operator_menu)
        self.menu.addSeparator()
        (
            self.channel_menu2,
            self.variable_menu2,
            operator_menu2,
            self.function_menu2,
        ), __ = variables_handling.get_menus(self.append_variable, "Append")
        self.menu.addMenu(self.channel_menu2)
        self.menu.addMenu(self.variable_menu2)
        self.menu.addMenu(self.function_menu2)
        self.menu.addMenu(operator_menu2)

        # Connect the search bar to the filtering function
        search_bar.textChanged.connect(self.filter_actions)

        self.menu.exec_(self.mapToGlobal(pos))

    def filter_actions(self, query):
        for self.menu in [
            self.channel_menu,
            self.variable_menu,
            self.function_menu,
            self.channel_menu2,
            self.variable_menu2,
            self.function_menu2,
        ]:
            for action in self.menu.actions():
                action.setVisible(query.lower() in action.text().lower())

    def append_variable(self, val):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        val :


        Returns
        -------

        """
        ind = self.tableWidget_channels.selectedIndexes()[0]
        item = self.tableWidget_channels.itemFromIndex(ind)
        text = item.text()
        item.setText(f"{text}{val}")

    def insert_variable(self, val):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        val :


        Returns
        -------

        """
        ind = self.tableWidget_channels.selectedIndexes()[0]
        item = self.tableWidget_channels.itemFromIndex(ind)
        item.setText(f"{val}")

    def check_change(self, pos):
        """

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        c = pos.column()
        if c != 0:
            return
        r = pos.row()
        item = self.tableWidget_channels.item(r, c)
        if item.checkState() != Qt.CheckState.Unchecked:
            color = variables_handling.get_color("blue")
        else:
            color = variables_handling.get_color("white")
        item.setBackground(QBrush(color))
        self.tableWidget_channels.item(r, c + 1).setBackground(QBrush(color))

    def change_search(self):
        """ """
        self.update_info()
        self.build_channels_table()

    def get_info(self):
        """ """
        self.update_info()
        return self.info_dict

    def check_string(self, item):
        """If an element is part of the checkstrings, the item becomes
        green if valid, red otherwise and white if empty.

        Parameters
        ----------
        item :


        Returns
        -------

        """
        if item.column() not in self.checkstrings:
            return
        if (
            self.tableWidget_channels.item(item.row(), 0).checkState()
            == Qt.CheckState.Unchecked
            and item.text() == ""
        ):
            color = variables_handling.get_color("white")
        elif variables_handling.check_eval(item.text()):
            color = variables_handling.get_color("green")
        else:
            color = variables_handling.get_color("red")
        item.setBackground(QBrush(color))

    def update_info(self):
        """ """
        channel_list = self.info_dict["channel"]
        if "value"  in self.info_dict:
            self.value_list = self.info_dict["value"].copy()
        for i in range(self.tableWidget_channels.rowCount()):
            name = self.tableWidget_channels.item(i, 1).text()
            if (
                name not in channel_list
                and self.tableWidget_channels.item(i, 0).checkState()
                != Qt.CheckState.Unchecked
            ):
                channel_list.append(name)
            elif (
                name in channel_list
                and self.tableWidget_channels.item(i, 0).checkState()
                == Qt.CheckState.Unchecked
            ):
                channel_list.remove(name)
            if name in channel_list:
                n = channel_list.index(name)
                for j, lab in enumerate(self.headerLabels[2:]):
                    while len(self.info_dict[lab]) < n + 1:
                        self.info_dict[lab].append(None)
                    item = self.tableWidget_channels.item(i, 2 + j)
                    t = item.text()
                    if not t:
                        raise Exception(
                            f"You need to enter a value for channel {name}!"
                        )
                    if t == 'None':
                        t = self.value_list[n]
                    self.info_dict[lab][n] = t
        rems = []
        for channel in channel_list:
            if channel not in self.channels:
                rems.append(channel)
        for channel in rems:
            channel_list.remove(channel)

    def build_channels_table(self):
        """ """
        self.tableWidget_channels.clear()
        self.tableWidget_channels.setColumnCount(len(self.headerLabels))
        self.tableWidget_channels.setRowCount(0)
        self.tableWidget_channels.setHorizontalHeaderLabels(self.headerLabels)
        searchtext = self.lineEdit_search.text()
        channel_list = self.info_dict["channel"]
        n = 0
        for i, channel in enumerate(sorted(self.channels, key=lambda x: x.lower())):
            if searchtext.lower() not in channel.lower() or (
                not isinstance(self.channels, list)
                and self.only_output
                and not self.channels[channel].output
            ):
                continue
            metadata = ""
            if not isinstance(self.channels, list):
                metadata = self.channels[channel].get_meta_str()
            self.tableWidget_channels.setRowCount(n + 1)
            item = QTableWidgetItem()
            item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            if channel in channel_list:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            if metadata:
                item.setToolTip(metadata)
            self.tableWidget_channels.setItem(n, 0, item)
            item = QTableWidgetItem(channel)
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            if metadata:
                item.setToolTip(metadata)
            self.tableWidget_channels.setItem(n, 1, item)
            pos = self.tableWidget_channels.model().createIndex(n, 0)
            self.check_change(pos)
            vals = []
            if channel in self.info_dict["channel"]:
                n_chan = self.info_dict["channel"].index(channel)
                for lab in self.headerLabels[2:]:
                    if n_chan < len(self.info_dict[lab]):  # Check if index is within range
                        vals.append(str(self.info_dict[lab][n_chan]))
                    else:
                        vals.append("")  # Append empty string if index is out of range
            for j in range(len(self.headerLabels[2:])):
                item = QTableWidgetItem(vals[j] if vals else "")
                self.tableWidget_channels.setItem(n, j + 2, item)
                self.check_string(item)
                if metadata:
                    item.setToolTip(metadata)
            n += 1
        self.tableWidget_channels.resizeColumnsToContents()


class Call_Functions_Table(QWidget):
    def __init__(
        self, parent=None, headerLabels=None, info_dict=None, title="", functions=None
    ):
        super().__init__(parent)
        self.functions = functions or variables_handling.get_non_channel_functions()
        self.headerLabels = headerLabels or []
        self.info_dict = info_dict or {}
        if "functions" not in self.info_dict:
            self.info_dict["functions"] = []

        layout = QGridLayout()
        self.tableWidget_functions = QTableWidget()
        self.tableWidget_functions.setHorizontalHeaderLabels(self.headerLabels)
        label_search = QLabel("Search:")
        self.tableWidget_functions.clicked.connect(
            self.tableWidget_functions.resizeColumnsToContents
        )
        self.lineEdit_search = QLineEdit()
        self.lineEdit_search.textChanged.connect(self.change_search)

        self.setLayout(layout)
        if title:
            title_label = QLabel(title)
            layout.addWidget(title_label, 0, 0, 1, 2)
            font = QFont()
            font.setBold(True)
            title_label.setStyleSheet("font-size: 9pt")
            title_label.setFont(font)
        layout.addWidget(label_search, 1, 0)
        layout.addWidget(self.lineEdit_search, 1, 1)
        layout.addWidget(self.tableWidget_functions, 2, 0, 1, 2)
        layout.setContentsMargins(0, 0, 0, 0)
        self.build_table()
        self.tableWidget_functions.clicked.connect(self.check_change)

    def check_change(self, pos):
        """ """
        c = pos.column()
        if c != 0:
            return
        r = pos.row()
        item = self.tableWidget_functions.item(r, c)
        if item.checkState() != Qt.CheckState.Unchecked:
            color = variables_handling.get_color("blue")
        else:
            color = variables_handling.get_color("white")
        item.setBackground(QBrush(color))
        self.tableWidget_functions.item(r, c + 1).setBackground(QBrush(color))

    def change_search(self):
        """ """
        self.update_info()
        self.build_table()

    def get_info(self):
        """ """
        self.update_info()
        return self.info_dict

    def update_info(self):
        """ """
        func_list = self.info_dict["functions"]
        for k in self.info_dict:
            if k != "functions":
                self.info_dict[k].clear()
        for i in range(self.tableWidget_functions.rowCount()):
            name = self.tableWidget_functions.item(i, 1).text()
            if (
                name not in func_list
                and self.tableWidget_functions.item(i, 0).checkState()
                != Qt.CheckState.Unchecked
            ):
                func_list.append(name)
            elif (
                name in func_list
                and self.tableWidget_functions.item(i, 0).checkState()
                == Qt.CheckState.Unchecked
            ):
                func_list.remove(name)
        rems = []
        for func in func_list:
            if func not in self.functions:
                rems.append(func)
        for func in rems:
            func_list.remove(func)

    def build_table(self):
        """ """
        self.tableWidget_functions.clear()
        self.tableWidget_functions.setColumnCount(len(self.headerLabels))
        self.tableWidget_functions.setRowCount(0)
        self.tableWidget_functions.setHorizontalHeaderLabels(self.headerLabels)
        searchtext = self.lineEdit_search.text()
        func_list = self.info_dict["functions"]
        n = 0
        for i, func in enumerate(sorted(self.functions, key=lambda x: x.lower())):
            if searchtext.lower() not in func.lower() or not isinstance(
                self.functions, list
            ):
                continue
            self.tableWidget_functions.setRowCount(n + 1)
            item = QTableWidgetItem()
            item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            if func in func_list:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            self.tableWidget_functions.setItem(n, 0, item)
            item = QTableWidgetItem(func)
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            self.tableWidget_functions.setItem(n, 1, item)
            pos = self.tableWidget_functions.model().createIndex(n, 0)
            self.check_change(pos)
            vals = []
            if func in self.info_dict["functions"]:
                n_chan = self.info_dict["functions"].index(func)
                for lab in self.headerLabels[2:]:
                    vals.append(str(self.info_dict[lab][n_chan]))
            for j in range(len(self.headerLabels[2:])):
                item = QTableWidgetItem(vals[j] if vals else "")
                self.tableWidget_functions.setItem(n, j + 2, item)
            n += 1
        self.tableWidget_functions.resizeColumnsToContents()
